from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import tensorflow as tf


def gelu(x):
    """Gaussian Error Linear Unit.

  This is a smoother version of the RELU.
  Original paper: https://arxiv.org/abs/1606.08415
  Args:
    x: float Tensor to perform activation.

  Returns:
    `x` with the GELU activation applied.
  """
    cdf = 0.5 * (
        1.0 + tf.tanh((np.sqrt(2 / np.pi) * (x + 0.044715 * tf.pow(x, 3))))
    )
    return x * cdf


def embedding_lookup(
    x,
    n_token,
    d_embed,
    hidden_size,
    initializer,
    use_tpu = True,
    scope = 'embedding',
    reuse = None,
    dtype = tf.float32,
):
    """TPU and GPU embedding_lookup function."""
    with tf.variable_scope(scope, reuse = reuse):
        lookup_table = tf.get_variable(
            'lookup_table',
            [n_token, d_embed],
            dtype = dtype,
            initializer = initializer,
        )

        project_variable = tf.get_variable(  # [embedding_size, hidden_size]
            name = 'lookup_table_2',
            shape = [d_embed, hidden_size],
            initializer = initializer,
        )

        if use_tpu:
            one_hot_idx = tf.one_hot(x, n_token, dtype = dtype)
            if one_hot_idx.shape.ndims == 2:
                return (
                    tf.einsum('in,nd->id', one_hot_idx, lookup_table),
                    lookup_table,
                )
            else:
                return (
                    tf.einsum('ibn,nd->ibd', one_hot_idx, lookup_table),
                    lookup_table,
                )
        else:
            output_middle = tf.nn.embedding_lookup(lookup_table, x)
            output = tf.matmul(output_middle, project_variable)
            return output, lookup_table, project_variable


def positional_embedding(pos_seq, inv_freq, bsz = None):
    sinusoid_inp = tf.einsum('i,d->id', pos_seq, inv_freq)
    pos_emb = tf.concat([tf.sin(sinusoid_inp), tf.cos(sinusoid_inp)], -1)
    pos_emb = pos_emb[:, None, :]

    if bsz is not None:
        pos_emb = tf.tile(pos_emb, [1, bsz, 1])

    return pos_emb


def positionwise_ffn(
    inp,
    d_model,
    d_inner,
    dropout,
    kernel_initializer,
    activation_type = 'relu',
    scope = 'ff',
    is_training = True,
    reuse = None,
):
    """Position-wise Feed-forward Network."""
    if activation_type == 'relu':
        activation = tf.nn.relu
    elif activation_type == 'gelu':
        activation = gelu
    else:
        raise ValueError(
            'Unsupported activation type {}'.format(activation_type)
        )

    output = inp
    with tf.variable_scope(scope, reuse = reuse):
        output = tf.layers.dense(
            output,
            d_inner,
            activation = activation,
            kernel_initializer = kernel_initializer,
            name = 'layer_1',
        )
        output = tf.layers.dropout(
            output, dropout, training = is_training, name = 'drop_1'
        )
        output = tf.layers.dense(
            output,
            d_model,
            kernel_initializer = kernel_initializer,
            name = 'layer_2',
        )
        output = tf.layers.dropout(
            output, dropout, training = is_training, name = 'drop_2'
        )
        output = tf.contrib.layers.layer_norm(
            output + inp, begin_norm_axis = -1, scope = 'LayerNorm'
        )
    return output


def head_projection(h, d_model, n_head, d_head, kernel_initializer, name):
    """Project hidden states to a specific head with a 4D-shape."""
    proj_weight = tf.get_variable(
        '{}/kernel'.format(name),
        [d_model, n_head, d_head],
        dtype = h.dtype,
        initializer = kernel_initializer,
    )
    head = tf.einsum('ibh,hnd->ibnd', h, proj_weight)

    return head


def post_attention(
    h,
    attn_vec,
    d_model,
    n_head,
    d_head,
    dropout,
    is_training,
    kernel_initializer,
    residual = True,
):
    """Post-attention processing."""
    # post-attention projection (back to `d_model`)
    proj_o = tf.get_variable(
        'o/kernel',
        [d_model, n_head, d_head],
        dtype = h.dtype,
        initializer = kernel_initializer,
    )
    attn_out = tf.einsum('ibnd,hnd->ibh', attn_vec, proj_o)

    attn_out = tf.layers.dropout(attn_out, dropout, training = is_training)
    if residual:
        output = tf.contrib.layers.layer_norm(
            attn_out + h, begin_norm_axis = -1, scope = 'LayerNorm'
        )
    else:
        output = tf.contrib.layers.layer_norm(
            attn_out, begin_norm_axis = -1, scope = 'LayerNorm'
        )

    return output


def abs_attn_core(
    q_head, k_head, v_head, attn_mask, dropatt, is_training, scale
):
    """Core absolute positional attention operations."""

    attn_score = tf.einsum('ibnd,jbnd->ijbn', q_head, k_head)
    attn_score *= scale
    if attn_mask is not None:
        attn_score = attn_score - 1e30 * attn_mask

    # attention probability
    attn_prob = tf.nn.softmax(attn_score, 1)
    attn_prob = tf.layers.dropout(attn_prob, dropatt, training = is_training)

    # attention output
    attn_vec = tf.einsum('ijbn,jbnd->ibnd', attn_prob, v_head)

    return attn_vec


def rel_attn_core(
    q_head,
    k_head_h,
    v_head_h,
    k_head_r,
    seg_embed,
    seg_mat,
    r_w_bias,
    r_r_bias,
    r_s_bias,
    attn_mask,
    dropatt,
    is_training,
    scale,
):
    """Core relative positional attention operations."""

    # content based attention score
    ac = tf.einsum('ibnd,jbnd->ijbn', q_head + r_w_bias, k_head_h)

    # position based attention score
    bd = tf.einsum('ibnd,jbnd->ijbn', q_head + r_r_bias, k_head_r)
    bd = rel_shift(bd, klen = tf.shape(ac)[1])

    # segment based attention score
    if seg_mat is None:
        ef = 0
    else:
        ef = tf.einsum('ibnd,snd->ibns', q_head + r_s_bias, seg_embed)
        ef = tf.einsum('ijbs,ibns->ijbn', seg_mat, ef)

    # merge attention scores and perform masking
    attn_score = (ac + bd + ef) * scale
    if attn_mask is not None:
        # attn_score = attn_score * (1 - attn_mask) - 1e30 * attn_mask
        attn_score = attn_score - 1e30 * attn_mask

    # attention probability
    attn_prob = tf.nn.softmax(attn_score, 1)
    attn_prob = tf.layers.dropout(attn_prob, dropatt, training = is_training)

    # attention output
    attn_vec = tf.einsum('ijbn,jbnd->ibnd', attn_prob, v_head_h)

    return attn_vec


def rel_shift(x, klen = -1):
    """perform relative shift to form the relative attention score."""
    x_size = tf.shape(x)

    x = tf.reshape(x, [x_size[1], x_size[0], x_size[2], x_size[3]])
    x = tf.slice(x, [1, 0, 0, 0], [-1, -1, -1, -1])
    x = tf.reshape(x, [x_size[0], x_size[1] - 1, x_size[2], x_size[3]])
    x = tf.slice(x, [0, 0, 0, 0], [-1, klen, -1, -1])

    return x


def _create_mask(qlen, mlen, dtype = tf.float32, same_length = False):
    """create causal attention mask."""
    attn_mask = tf.ones([qlen, qlen], dtype = dtype)
    mask_u = tf.matrix_band_part(attn_mask, 0, -1)
    mask_dia = tf.matrix_band_part(attn_mask, 0, 0)
    attn_mask_pad = tf.zeros([qlen, mlen], dtype = dtype)
    ret = tf.concat([attn_mask_pad, mask_u - mask_dia], 1)
    if same_length:
        mask_l = tf.matrix_band_part(attn_mask, -1, 0)
        ret = tf.concat([ret[:, :qlen] + mask_l - mask_dia, ret[:, qlen:]], 1)

    return ret


def _cache_mem(curr_out, prev_mem, mem_len, reuse_len = None):
    """cache hidden states into memory."""
    if mem_len is None or mem_len == 0:
        return None
    else:
        if reuse_len is not None and reuse_len > 0:
            curr_out = curr_out[:reuse_len]

        if prev_mem is None:
            new_mem = curr_out[-mem_len:]
        else:
            new_mem = tf.concat([prev_mem, curr_out], 0)[-mem_len:]

    return tf.stop_gradient(new_mem)


def relative_positional_encoding(
    qlen, klen, d_model, clamp_len, attn_type, bi_data, bsz = None, dtype = None
):
    """create relative positional encoding."""
    freq_seq = tf.range(0, d_model, 2.0)
    if dtype is not None and dtype != tf.float32:
        freq_seq = tf.cast(freq_seq, dtype = dtype)
    inv_freq = 1 / (10000 ** (freq_seq / d_model))

    if attn_type == 'bi':
        # beg, end = klen - 1, -qlen
        beg, end = klen, -qlen
    elif attn_type == 'uni':
        # beg, end = klen - 1, -1
        beg, end = klen, -1
    else:
        raise ValueError('Unknown `attn_type` {}.'.format(attn_type))

    if bi_data:
        fwd_pos_seq = tf.range(beg, end, -1.0)
        bwd_pos_seq = tf.range(-beg, -end, 1.0)

        if dtype is not None and dtype != tf.float32:
            fwd_pos_seq = tf.cast(fwd_pos_seq, dtype = dtype)
            bwd_pos_seq = tf.cast(bwd_pos_seq, dtype = dtype)

        if clamp_len > 0:
            fwd_pos_seq = tf.clip_by_value(fwd_pos_seq, -clamp_len, clamp_len)
            bwd_pos_seq = tf.clip_by_value(bwd_pos_seq, -clamp_len, clamp_len)

        tf.logging.info('bsz here', bsz)
        if bsz is not None:
            # With bi_data, the batch size should be divisible by 2.
            # assert bsz % 2 == 0
            fwd_pos_emb = positional_embedding(fwd_pos_seq, inv_freq, bsz // 2)
            bwd_pos_emb = positional_embedding(bwd_pos_seq, inv_freq, bsz // 2)
        else:
            fwd_pos_emb = positional_embedding(fwd_pos_seq, inv_freq)
            bwd_pos_emb = positional_embedding(bwd_pos_seq, inv_freq)

        pos_emb = tf.concat([fwd_pos_emb, bwd_pos_emb], axis = 1)
    else:
        fwd_pos_seq = tf.range(beg, end, -1.0)
        if dtype is not None and dtype != tf.float32:
            fwd_pos_seq = tf.cast(fwd_pos_seq, dtype = dtype)
        if clamp_len > 0:
            fwd_pos_seq = tf.clip_by_value(fwd_pos_seq, -clamp_len, clamp_len)
        pos_emb = positional_embedding(fwd_pos_seq, inv_freq, bsz)

    return pos_emb


def multihead_attn(
    q,
    k,
    v,
    attn_mask,
    d_model,
    n_head,
    d_head,
    dropout,
    dropatt,
    is_training,
    kernel_initializer,
    residual = True,
    scope = 'abs_attn',
    reuse = None,
):
    """Standard multi-head attention with absolute positional embedding."""

    scale = 1 / (d_head ** 0.5)
    with tf.variable_scope(scope, reuse = reuse):
        # attention heads
        q_head = head_projection(
            q, d_model, n_head, d_head, kernel_initializer, 'q'
        )
        k_head = head_projection(
            k, d_model, n_head, d_head, kernel_initializer, 'k'
        )
        v_head = head_projection(
            v, d_model, n_head, d_head, kernel_initializer, 'v'
        )

        # attention vector
        attn_vec = abs_attn_core(
            q_head, k_head, v_head, attn_mask, dropatt, is_training, scale
        )

        # post processing
        output = post_attention(
            v,
            attn_vec,
            d_model,
            n_head,
            d_head,
            dropout,
            is_training,
            kernel_initializer,
            residual,
        )

    return output


def rel_multihead_attn(
    h,
    r,
    r_w_bias,
    r_r_bias,
    seg_mat,
    r_s_bias,
    seg_embed,
    attn_mask,
    mems,
    d_model,
    n_head,
    d_head,
    dropout,
    dropatt,
    is_training,
    kernel_initializer,
    scope = 'rel_attn',
    reuse = None,
):
    """Multi-head attention with relative positional encoding."""

    scale = 1 / (d_head ** 0.5)
    with tf.variable_scope(scope, reuse = reuse):
        if mems is not None and mems.shape.ndims > 1:
            cat = tf.concat([mems, h], 0)
        else:
            cat = h

        # content heads
        q_head_h = head_projection(
            h, d_model, n_head, d_head, kernel_initializer, 'q'
        )
        k_head_h = head_projection(
            cat, d_model, n_head, d_head, kernel_initializer, 'k'
        )
        v_head_h = head_projection(
            cat, d_model, n_head, d_head, kernel_initializer, 'v'
        )

        # positional heads
        k_head_r = head_projection(
            r, d_model, n_head, d_head, kernel_initializer, 'r'
        )

        # core attention ops
        attn_vec = rel_attn_core(
            q_head_h,
            k_head_h,
            v_head_h,
            k_head_r,
            seg_embed,
            seg_mat,
            r_w_bias,
            r_r_bias,
            r_s_bias,
            attn_mask,
            dropatt,
            is_training,
            scale,
        )

        # post processing
        output = post_attention(
            h,
            attn_vec,
            d_model,
            n_head,
            d_head,
            dropout,
            is_training,
            kernel_initializer,
        )

    return output


def two_stream_rel_attn(
    h,
    g,
    r,
    mems,
    r_w_bias,
    r_r_bias,
    seg_mat,
    r_s_bias,
    seg_embed,
    attn_mask_h,
    attn_mask_g,
    target_mapping,
    d_model,
    n_head,
    d_head,
    dropout,
    dropatt,
    is_training,
    kernel_initializer,
    scope = 'rel_attn',
):
    """Two-stream attention with relative positional encoding."""

    scale = 1 / (d_head ** 0.5)
    with tf.variable_scope(scope, reuse = False):

        # content based attention score
        if mems is not None and mems.shape.ndims > 1:
            cat = tf.concat([mems, h], 0)
        else:
            cat = h

        # content-based key head
        k_head_h = head_projection(
            cat, d_model, n_head, d_head, kernel_initializer, 'k'
        )

        # content-based value head
        v_head_h = head_projection(
            cat, d_model, n_head, d_head, kernel_initializer, 'v'
        )

        # position-based key head
        k_head_r = head_projection(
            r, d_model, n_head, d_head, kernel_initializer, 'r'
        )

        ##### h-stream
        # content-stream query head
        q_head_h = head_projection(
            h, d_model, n_head, d_head, kernel_initializer, 'q'
        )

        # core attention ops
        attn_vec_h = rel_attn_core(
            q_head_h,
            k_head_h,
            v_head_h,
            k_head_r,
            seg_embed,
            seg_mat,
            r_w_bias,
            r_r_bias,
            r_s_bias,
            attn_mask_h,
            dropatt,
            is_training,
            scale,
        )

        # post processing
        output_h = post_attention(
            h,
            attn_vec_h,
            d_model,
            n_head,
            d_head,
            dropout,
            is_training,
            kernel_initializer,
        )

    with tf.variable_scope(scope, reuse = True):
        ##### g-stream
        # query-stream query head
        q_head_g = head_projection(
            g, d_model, n_head, d_head, kernel_initializer, 'q'
        )

        # core attention ops
        if target_mapping is not None:
            q_head_g = tf.einsum('mbnd,mlb->lbnd', q_head_g, target_mapping)
            attn_vec_g = rel_attn_core(
                q_head_g,
                k_head_h,
                v_head_h,
                k_head_r,
                seg_embed,
                seg_mat,
                r_w_bias,
                r_r_bias,
                r_s_bias,
                attn_mask_g,
                dropatt,
                is_training,
                scale,
            )
            attn_vec_g = tf.einsum('lbnd,mlb->mbnd', attn_vec_g, target_mapping)
        else:
            attn_vec_g = rel_attn_core(
                q_head_g,
                k_head_h,
                v_head_h,
                k_head_r,
                seg_embed,
                seg_mat,
                r_w_bias,
                r_r_bias,
                r_s_bias,
                attn_mask_g,
                dropatt,
                is_training,
                scale,
            )

        # post processing
        output_g = post_attention(
            g,
            attn_vec_g,
            d_model,
            n_head,
            d_head,
            dropout,
            is_training,
            kernel_initializer,
        )

        return output_h, output_g


def transformer_xl(
    inp_k,
    n_token,
    n_layer,
    d_model,
    n_head,
    d_head,
    d_inner,
    dropout,
    dropatt,
    attn_type,
    bi_data,
    initializer,
    is_training,
    mem_len = None,
    inp_q = None,
    mems = None,
    same_length = False,
    clamp_len = -1,
    untie_r = False,
    use_tpu = True,
    input_mask = None,
    perm_mask = None,
    seg_id = None,
    reuse_len = None,
    ff_activation = 'relu',
    target_mapping = None,
    use_bfloat16 = False,
    scope = 'transformer',
    **kwargs
):
    """
    Defines a Transformer-XL computation graph with additional
    support for XLNet.

    Args:

    inp_k: int32 Tensor in shape [len, bsz], the input token IDs.
    seg_id: int32 Tensor in shape [len, bsz], the input segment IDs.
    input_mask: float32 Tensor in shape [len, bsz], the input mask.
      0 for real tokens and 1 for padding.
    mems: a list of float32 Tensors in shape [mem_len, bsz, d_model], memory
      from previous batches. The length of the list equals n_layer.
      If None, no memory is used.
    perm_mask: float32 Tensor in shape [len, len, bsz].
      If perm_mask[i, j, k] = 0, i attend to j in batch k;
      if perm_mask[i, j, k] = 1, i does not attend to j in batch k.
      If None, each position attends to all the others.
    target_mapping: float32 Tensor in shape [num_predict, len, bsz].
      If target_mapping[i, j, k] = 1, the i-th predict in batch k is
      on the j-th token.
      Only used during pretraining for partial prediction.
      Set to None during finetuning.
    inp_q: float32 Tensor in shape [len, bsz].
      1 for tokens with losses and 0 for tokens without losses.
      Only used during pretraining for two-stream attention.
      Set to None during finetuning.

    n_layer: int, the number of layers.
    d_model: int, the hidden size.
    n_head: int, the number of attention heads.
    d_head: int, the dimension size of each attention head.
    d_inner: int, the hidden size in feed-forward layers.
    ff_activation: str, "relu" or "gelu".
    untie_r: bool, whether to untie the biases in attention.
    n_token: int, the vocab size.

    is_training: bool, whether in training mode.
    use_tpu: bool, whether TPUs are used.
    use_bfloat16: bool, use bfloat16 instead of float32.
    dropout: float, dropout rate.
    dropatt: float, dropout rate on attention probabilities.
    init: str, the initialization scheme, either "normal" or "uniform".
    init_range: float, initialize the parameters with a uniform distribution
      in [-init_range, init_range]. Only effective when init="uniform".
    init_std: float, initialize the parameters with a normal distribution
      with mean 0 and stddev init_std. Only effective when init="normal".
    mem_len: int, the number of tokens to cache.
    reuse_len: int, the number of tokens in the currect batch to be cached
      and reused in the future.
    bi_data: bool, whether to use bidirectional input pipeline.
      Usually set to True during pretraining and False during finetuning.
    clamp_len: int, clamp all relative distances larger than clamp_len.
      -1 means no clamping.
    same_length: bool, whether to use the same attention length for each token.
    summary_type: str, "last", "first", "mean", or "attn". The method
      to pool the input to get a vector representation.
    initializer: A tf initializer.
    scope: scope name for the computation graph.
  """
    tf.logging.info('memory input {}'.format(mems))
    tf_float = tf.bfloat16 if use_bfloat16 else tf.float32
    tf.logging.info('Use float type {}'.format(tf_float))

    new_mems = []
    with tf.variable_scope(scope):
        if untie_r:
            r_w_bias = tf.get_variable(
                'r_w_bias',
                [n_layer, n_head, d_head],
                dtype = tf_float,
                initializer = initializer,
            )
            r_r_bias = tf.get_variable(
                'r_r_bias',
                [n_layer, n_head, d_head],
                dtype = tf_float,
                initializer = initializer,
            )
        else:
            r_w_bias = tf.get_variable(
                'r_w_bias',
                [n_head, d_head],
                dtype = tf_float,
                initializer = initializer,
            )
            r_r_bias = tf.get_variable(
                'r_r_bias',
                [n_head, d_head],
                dtype = tf_float,
                initializer = initializer,
            )

        bsz = tf.shape(inp_k)[1]
        qlen = tf.shape(inp_k)[0]
        mlen = tf.shape(mems[0])[0] if mems is not None else 0
        klen = mlen + qlen

        ##### Attention mask
        # causal attention mask
        if attn_type == 'uni':
            attn_mask = _create_mask(qlen, mlen, tf_float, same_length)
            attn_mask = attn_mask[:, :, None, None]
        elif attn_type == 'bi':
            attn_mask = None
        else:
            raise ValueError('Unsupported attention type: {}'.format(attn_type))

        # data mask: input mask & perm mask
        if input_mask is not None and perm_mask is not None:
            data_mask = input_mask[None] + perm_mask
        elif input_mask is not None and perm_mask is None:
            data_mask = input_mask[None]
        elif input_mask is None and perm_mask is not None:
            data_mask = perm_mask
        else:
            data_mask = None

        if data_mask is not None:
            # all mems can be attended to
            mems_mask = tf.zeros(
                [tf.shape(data_mask)[0], mlen, bsz], dtype = tf_float
            )
            data_mask = tf.concat([mems_mask, data_mask], 1)
            if attn_mask is None:
                attn_mask = data_mask[:, :, :, None]
            else:
                attn_mask += data_mask[:, :, :, None]

        if attn_mask is not None:
            attn_mask = tf.cast(attn_mask > 0, dtype = tf_float)

        if attn_mask is not None:
            non_tgt_mask = -tf.eye(qlen, dtype = tf_float)
            non_tgt_mask = tf.concat(
                [tf.zeros([qlen, mlen], dtype = tf_float), non_tgt_mask],
                axis = -1,
            )
            non_tgt_mask = tf.cast(
                (attn_mask + non_tgt_mask[:, :, None, None]) > 0,
                dtype = tf_float,
            )
        else:
            non_tgt_mask = None

        ##### Word embedding
        word_emb_k, lookup_table, lookup_table_2 = embedding_lookup(
            x = inp_k,
            n_token = n_token,
            d_embed = 128,
            hidden_size = d_model,
            initializer = initializer,
            use_tpu = use_tpu,
            dtype = tf_float,
            scope = 'word_embedding',
        )

        if inp_q is not None:
            with tf.variable_scope('mask_emb'):
                mask_emb = tf.get_variable(
                    'mask_emb', [1, 1, d_model], dtype = tf_float
                )
                if target_mapping is not None:
                    word_emb_q = tf.tile(
                        mask_emb, [tf.shape(target_mapping)[0], bsz, 1]
                    )
                else:
                    inp_q_ext = inp_q[:, :, None]
                    word_emb_q = (
                        inp_q_ext * mask_emb + (1 - inp_q_ext) * word_emb_k
                    )
        output_h = tf.layers.dropout(
            word_emb_k, dropout, training = is_training
        )
        if inp_q is not None:
            output_g = tf.layers.dropout(
                word_emb_q, dropout, training = is_training
            )

        ##### Segment embedding
        if seg_id is not None:
            if untie_r:
                r_s_bias = tf.get_variable(
                    'r_s_bias',
                    [n_layer, n_head, d_head],
                    dtype = tf_float,
                    initializer = initializer,
                )
            else:
                # default case (tie)
                r_s_bias = tf.get_variable(
                    'r_s_bias',
                    [n_head, d_head],
                    dtype = tf_float,
                    initializer = initializer,
                )

            seg_embed = tf.get_variable(
                'seg_embed',
                [n_layer, 2, n_head, d_head],
                dtype = tf_float,
                initializer = initializer,
            )

            # Convert `seg_id` to one-hot `seg_mat`
            mem_pad = tf.zeros([mlen, bsz], dtype = tf.int32)
            cat_ids = tf.concat([mem_pad, seg_id], 0)

            # `1` indicates not in the same segment [qlen x klen x bsz]
            seg_mat = tf.cast(
                tf.logical_not(tf.equal(seg_id[:, None], cat_ids[None, :])),
                tf.int32,
            )
            seg_mat = tf.one_hot(seg_mat, 2, dtype = tf_float)
        else:
            seg_mat = None

        ##### Positional encoding
        pos_emb = relative_positional_encoding(
            qlen,
            klen,
            d_model,
            clamp_len,
            attn_type,
            bi_data,
            bsz = bsz,
            dtype = tf_float,
        )
        pos_emb = tf.layers.dropout(pos_emb, dropout, training = is_training)

        ##### Attention layers
        if mems is None:
            mems = [None] * n_layer

        name_variable_scope = 'layer_shared'

        for i in range(n_layer):
            # cache new mems
            new_mems.append(_cache_mem(output_h, mems[i], mem_len, reuse_len))

            # segment bias
            if seg_id is None:
                r_s_bias_i = None
                seg_embed_i = None
            else:
                r_s_bias_i = r_s_bias if not untie_r else r_s_bias[i]
                seg_embed_i = seg_embed[i]

            with tf.variable_scope(
                name_variable_scope, reuse = True if i > 0 else False
            ):
                if inp_q is not None:
                    output_h, output_g = two_stream_rel_attn(
                        h = output_h,
                        g = output_g,
                        r = pos_emb,
                        r_w_bias = r_w_bias if not untie_r else r_w_bias[i],
                        r_r_bias = r_r_bias if not untie_r else r_r_bias[i],
                        seg_mat = seg_mat,
                        r_s_bias = r_s_bias_i,
                        seg_embed = seg_embed_i,
                        attn_mask_h = non_tgt_mask,
                        attn_mask_g = attn_mask,
                        mems = mems[i],
                        target_mapping = target_mapping,
                        d_model = d_model,
                        n_head = n_head,
                        d_head = d_head,
                        dropout = dropout,
                        dropatt = dropatt,
                        is_training = is_training,
                        kernel_initializer = initializer,
                    )
                    reuse = True
                else:
                    reuse = False

                    output_h = rel_multihead_attn(
                        h = output_h,
                        r = pos_emb,
                        r_w_bias = r_w_bias if not untie_r else r_w_bias[i],
                        r_r_bias = r_r_bias if not untie_r else r_r_bias[i],
                        seg_mat = seg_mat,
                        r_s_bias = r_s_bias_i,
                        seg_embed = seg_embed_i,
                        attn_mask = non_tgt_mask,
                        mems = mems[i],
                        d_model = d_model,
                        n_head = n_head,
                        d_head = d_head,
                        dropout = dropout,
                        dropatt = dropatt,
                        is_training = is_training,
                        kernel_initializer = initializer,
                        reuse = reuse,
                    )

                if inp_q is not None:
                    output_g = positionwise_ffn(
                        inp = output_g,
                        d_model = d_model,
                        d_inner = d_inner,
                        dropout = dropout,
                        kernel_initializer = initializer,
                        activation_type = ff_activation,
                        is_training = is_training,
                    )

                output_h = positionwise_ffn(
                    inp = output_h,
                    d_model = d_model,
                    d_inner = d_inner,
                    dropout = dropout,
                    kernel_initializer = initializer,
                    activation_type = ff_activation,
                    is_training = is_training,
                    reuse = reuse,
                )

        if inp_q is not None:
            output = tf.layers.dropout(
                output_g, dropout, training = is_training
            )
        else:
            output = tf.layers.dropout(
                output_h, dropout, training = is_training
            )

        return output, new_mems, lookup_table, lookup_table_2


def lm_loss(
    hidden,
    target,
    n_token,
    d_model,
    initializer,
    lookup_table = None,
    tie_weight = False,
    bi_data = True,
    use_tpu = False,
):
    """doc."""

    with tf.variable_scope('lm_loss'):
        if tie_weight:
            assert (
                lookup_table is not None
            ), 'lookup_table cannot be None for tie_weight'
            softmax_w = lookup_table
        else:
            softmax_w = tf.get_variable(
                'weight',
                [n_token, d_model],
                dtype = hidden.dtype,
                initializer = initializer,
            )

        softmax_b = tf.get_variable(
            'bias',
            [n_token],
            dtype = hidden.dtype,
            initializer = tf.zeros_initializer(),
        )

        logits = tf.einsum('ibd,nd->ibn', hidden, softmax_w) + softmax_b

        if use_tpu:
            one_hot_target = tf.one_hot(target, n_token, dtype = logits.dtype)
            loss = -tf.reduce_sum(
                tf.nn.log_softmax(logits) * one_hot_target, -1
            )
        else:
            loss = tf.nn.sparse_softmax_cross_entropy_with_logits(
                labels = target, logits = logits
            )

        return loss


def lm_accuracy(
    hidden,
    target,
    n_token,
    d_model,
    initializer,
    lookup_table = None,
    lookup_table_2 = None,
    tie_weight = False,
    bi_data = True,
    use_tpu = False,
):
    """doc."""

    with tf.variable_scope('lm_loss'):
        if tie_weight:
            assert (
                lookup_table is not None
            ), 'lookup_table cannot be None for tie_weight'
            softmax_w = lookup_table
        else:
            softmax_w = tf.get_variable(
                'weight',
                [n_token, d_model],
                dtype = hidden.dtype,
                initializer = initializer,
            )

        softmax_b = tf.get_variable(
            'bias',
            [n_token],
            dtype = hidden.dtype,
            initializer = tf.zeros_initializer(),
        )
        softmax_w = tf.matmul(softmax_w, lookup_table_2)

        logits = tf.einsum('ibd,nd->ibn', hidden, softmax_w) + softmax_b
        next_sentence_predictions = tf.argmax(
            logits, axis = -1, output_type = tf.int32
        )
        print(target, next_sentence_predictions)
        next_sentence_accuracy = tf.metrics.accuracy(
            labels = target, predictions = next_sentence_predictions
        )

        if use_tpu:
            one_hot_target = tf.one_hot(target, n_token, dtype = logits.dtype)
            loss = -tf.reduce_sum(
                tf.nn.log_softmax(logits) * one_hot_target, -1
            )
        else:
            loss = tf.nn.sparse_softmax_cross_entropy_with_logits(
                labels = target, logits = logits
            )

        return next_sentence_accuracy, loss


def summarize_sequence(
    summary_type,
    hidden,
    d_model,
    n_head,
    d_head,
    dropout,
    dropatt,
    input_mask,
    is_training,
    initializer,
    scope = None,
    reuse = None,
    use_proj = True,
):

    """
      Different classification tasks may not may not share the same parameters
      to summarize the sequence features.

      If shared, one can keep the `scope` to the default value `None`.
      Otherwise, one should specify a different `scope` for each task.
  """

    with tf.variable_scope(scope, 'sequnece_summary', reuse = reuse):
        if summary_type == 'last':
            summary = hidden[-1]
        elif summary_type == 'first':
            summary = hidden[0]
        elif summary_type == 'mean':
            summary = tf.reduce_mean(hidden, axis = 0)
        elif summary_type == 'attn':
            bsz = tf.shape(hidden)[1]

            summary_bias = tf.get_variable(
                'summary_bias',
                [d_model],
                dtype = hidden.dtype,
                initializer = initializer,
            )
            summary_bias = tf.tile(summary_bias[None, None], [1, bsz, 1])

            if input_mask is not None:
                input_mask = input_mask[None, :, :, None]

            summary = multihead_attn(
                summary_bias,
                hidden,
                hidden,
                input_mask,
                d_model,
                n_head,
                d_head,
                dropout,
                dropatt,
                is_training,
                initializer,
                residual = False,
            )
            summary = summary[0]
        else:
            raise ValueError('Unsupported summary type {}'.format(summary_type))

        # use another projection as in BERT
        if use_proj:
            summary = tf.layers.dense(
                summary,
                d_model,
                activation = tf.tanh,
                kernel_initializer = initializer,
                name = 'summary',
            )

        # dropout
        summary = tf.layers.dropout(
            summary, dropout, training = is_training, name = 'dropout'
        )

    return summary


def classification_loss(
    hidden,
    labels,
    n_class,
    initializer,
    scope,
    reuse = None,
    return_logits = False,
):
    """
      Different classification tasks should use different scope names to ensure
      different dense layers (parameters) are used to produce the logits.

      An exception will be in transfer learning, where one hopes to transfer
      the classification weights.
  """

    with tf.variable_scope(scope, reuse = reuse):
        logits = tf.layers.dense(
            hidden, n_class, kernel_initializer = initializer, name = 'logit'
        )

        one_hot_target = tf.one_hot(labels, n_class, dtype = hidden.dtype)
        loss = -tf.reduce_sum(tf.nn.log_softmax(logits) * one_hot_target, -1)

        if return_logits:
            return loss, logits

        return loss


def regression_loss(
    hidden, labels, initializer, scope, reuse = None, return_logits = False
):
    with tf.variable_scope(scope, reuse = reuse):
        logits = tf.layers.dense(
            hidden, 1, kernel_initializer = initializer, name = 'logit'
        )

        logits = tf.squeeze(logits, axis = -1)
        loss = tf.square(logits - labels)

        if return_logits:
            return loss, logits

        return loss



from scipy.stats.mstats import gmean, hmean, hdmedian
import numpy as np


def _most_common(l):
    return max(set(l), key = l.count)


def voting_stack(models, text):
    """
    Stacking for POS, Entities and Dependency models.

    Parameters
    ----------
    models: list
        list of models.
    text: str
        string to predict.

    Returns
    -------
    result: list
    """
    if not isinstance(models, list):
        raise ValueError('models must be a list')
    if not isinstance(text, str):
        raise ValueError('text must be a string')
    results, texts, votes, votes_indices, indices = [], [], [], [], []
    is_dependency = False
    for i in range(len(models)):
        if not 'predict' in dir(models[i]):
            raise ValueError('all models must able to predict')
        predicted = models[i].predict(text)
        if isinstance(predicted, tuple):
            is_dependency = True
            d, predicted, indexing = predicted
            indexing = np.array(indexing)
            indices.append(indexing[:, 1:2])

        predicted = np.array(predicted)
        results.append(predicted[:, 1:2])
        texts.append(predicted[:, 0])
    concatenated = np.concatenate(results, axis = 1)
    for row in concatenated:
        votes.append(_most_common(row.tolist()))
    if is_dependency:
        concatenated = np.concatenate(indices, axis = 1)
        for row in concatenated:
            votes_indices.append(_most_common(row.tolist()))
    output = list(map(lambda X: (X[0], X[1]), list(zip(texts[-1], votes))))
    if is_dependency:
        return (
            output,
            list(
                map(lambda X: (X[0], X[1]), list(zip(texts[-1], votes_indices)))
            ),
        )
    else:
        return output


dict_function = {
    'gmean': gmean,
    'hmean': hmean,
    'mean': np.mean,
    'min': np.amin,
    'max': np.amax,
    'median': hdmedian,
}


def predict_stack(models, strings, mode = 'gmean'):
    """
    Stacking for predictive models.

    Parameters
    ----------
    models: list
        list of models.
    strings: str or list of str
        strings to predict.
    mode : str, optional (default='gmean')
        Model architecture supported. Allowed values:

        * ``'gmean'`` - geometrical mean.
        * ``'hmean'`` - harmonic mean.
        * ``'mean'`` - mean.
        * ``'min'`` - min.
        * ``'max'`` - max.
        * ``'median'`` - Harrell-Davis median.


    Returns
    -------
    result: dict
    """
    if not isinstance(models, list):
        raise ValueError('models must be a list')
    if isinstance(strings, list):
        if not isinstance(strings[0], str):
            raise ValueError('input must be a list of strings or a string')
    else:
        if not isinstance(strings, str):
            raise ValueError('input must be a list of strings or a string')
    if isinstance(strings, str):
        strings = [strings]
    if not isinstance(mode, str):
        raise ValueError('mode must be a string')
    if mode.lower() not in dict_function:
        raise Exception(
            "mode not supported, only support ['gmean','hmean','mean','min','max','median']"
        )
    mode = dict_function[mode.lower()]

    for i in range(len(models)):
        if not 'predict_batch' in dir(models[i]):
            raise ValueError('all models must able to predict_batch')

    labels, results = None, []
    for i in range(len(models)):
        nested_results = []
        result = (
            models[i].predict_batch(strings)
            if models[i].predict.__defaults__ is None
            else models[i].predict_batch(strings, get_proba = True)
        )
        for r in result:
            l = list(r.keys())
            if not labels:
                labels = l
            else:
                if l != labels:
                    raise ValueError('domain classification must be same!')
            nested_results.append(list(r.values()))
        results.append(nested_results)
    results = mode(np.array(results), axis = 0)
    outputs = []
    for result in results:
        outputs.append({label: result[no] for no, label in enumerate(labels)})
    return outputs




# coding: utf-8

# In[1]:


import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter('ignore')

import tensorflow as tf
import numpy as np
from tqdm import tqdm
import re
import collections
import json
import os
from unidecode import unidecode


def batch_sequence(sentences, dictionary, maxlen = 50):
    np_array = np.zeros((len(sentences), maxlen), dtype = np.int32)
    for no_sentence, sentence in enumerate(sentences):
        current_no = 0
        for no, word in enumerate(sentence.split()[: maxlen - 2]):
            np_array[no_sentence, no] = dictionary.get(word, 1)
            current_no = no
        np_array[no_sentence, current_no + 1] = 3
    return np_array


class Attention:
    def __init__(self,hidden_size):
        self.hidden_size = hidden_size
        self.dense_layer = tf.layers.Dense(hidden_size)
        self.v = tf.random_normal([hidden_size],mean=0,stddev=1/np.sqrt(hidden_size))
        
    def score(self, hidden_tensor, encoder_outputs):
        energy = tf.nn.tanh(self.dense_layer(tf.concat([hidden_tensor,encoder_outputs],2)))
        energy = tf.transpose(energy,[0,2,1])
        batch_size = tf.shape(encoder_outputs)[0]
        v = tf.expand_dims(tf.tile(tf.expand_dims(self.v,0),[batch_size,1]),1)
        energy = tf.matmul(v,energy)
        return tf.squeeze(energy,1)
    
    def __call__(self, hidden, encoder_outputs):
        seq_len = tf.shape(encoder_outputs)[1]
        batch_size = tf.shape(encoder_outputs)[0]
        H = tf.tile(tf.expand_dims(hidden, 1),[1,seq_len,1])
        attn_energies = self.score(H,encoder_outputs)
        return tf.expand_dims(tf.nn.softmax(attn_energies),1)

class Model:
    def __init__(
        self,
        dict_size,
        size_layers,
        learning_rate,
        maxlen,
        num_blocks = 3,
    ):
        block_size = size_layers
        self.BEFORE = tf.placeholder(tf.int32,[None,maxlen])
        self.INPUT = tf.placeholder(tf.int32,[None,maxlen])
        self.AFTER = tf.placeholder(tf.int32,[None,maxlen])
        self.batch_size = tf.shape(self.INPUT)[0]
        self.output_layer = tf.layers.Dense(dict_size, name="output_layer")
        self.output_layer.build(size_layers)
        self.embeddings = tf.Variable(tf.random_uniform([dict_size, size_layers], -1, 1))
        embedded = tf.nn.embedding_lookup(self.embeddings, self.INPUT)
        self.attention = Attention(size_layers)

        def residual_block(x, size, rate, block, reuse = False):
            with tf.variable_scope(
                'block_%d_%d' % (block, rate), reuse = reuse
            ):
                attn_weights = self.attention(tf.reduce_sum(x,axis=1), x)
                conv_filter = tf.layers.conv1d(
                    attn_weights,
                    x.shape[2] // 4,
                    kernel_size = size,
                    strides = 1,
                    padding = 'same',
                    dilation_rate = rate,
                    activation = tf.nn.tanh,
                )
                conv_gate = tf.layers.conv1d(
                    x,
                    x.shape[2] // 4,
                    kernel_size = size,
                    strides = 1,
                    padding = 'same',
                    dilation_rate = rate,
                    activation = tf.nn.sigmoid,
                )
                out = tf.multiply(conv_filter, conv_gate)
                out = tf.layers.conv1d(
                    out,
                    block_size,
                    kernel_size = 1,
                    strides = 1,
                    padding = 'same',
                    activation = tf.nn.tanh,
                )
                return tf.add(x, out), out

        forward = tf.layers.conv1d(
            embedded, block_size, kernel_size = 1, strides = 1, padding = 'SAME'
        )
        zeros = tf.zeros_like(forward)
        for i in range(num_blocks):
            for r in [1, 2, 4, 8, 16]:
                forward, s = residual_block(
                    forward, size = 7, rate = r, block = i
                )
                zeros = tf.add(zeros, s)
        forward = tf.layers.conv1d(
            zeros,
            block_size,
            kernel_size = 1,
            strides = 1,
            padding = 'SAME',
            activation = tf.nn.tanh,
        )
        self.get_thought = tf.reduce_sum(forward,axis=1, name = 'logits')
        
        def decoder(labels, reuse):
            decoder_in = tf.nn.embedding_lookup(self.embeddings, labels)
            forward = tf.layers.conv1d(
                decoder_in, block_size, kernel_size = 1, strides = 1, padding = 'SAME'
            )
            zeros = tf.zeros_like(forward)
            for r in [8, 16, 24]:
                forward, s = residual_block(forward, size = 7, rate = r, block = 10, reuse = reuse)
                zeros = tf.add(zeros, s)
            return tf.layers.conv1d(
                zeros,
                block_size,
                kernel_size = 1,
                strides = 1,
                padding = 'SAME',
                activation = tf.nn.tanh,
            )
        
        fw_logits = decoder(self.AFTER, False)
        bw_logits = decoder(self.BEFORE, True)
        self.attention = tf.matmul(
            self.get_thought, tf.transpose(self.embeddings), name = 'attention'
        )
        self.loss = self.calculate_loss(fw_logits, self.AFTER) + self.calculate_loss(bw_logits, self.BEFORE)
        self.optimizer = tf.train.AdamOptimizer(learning_rate).minimize(self.loss)
    
    def calculate_loss(self, outputs, labels):
        mask = tf.cast(tf.sign(labels), tf.float32)
        logits = self.output_layer(outputs)
        return tf.contrib.seq2seq.sequence_loss(logits, labels, mask)


def counter_words(sentences):
    word_counter = collections.Counter()
    word_list = []
    num_lines, num_words = (0, 0)
    for i in sentences:
        words = re.findall('[\\w\']+|[;:\-\(\)&.,!?"]', i)
        word_counter.update(words)
        word_list.extend(words)
        num_lines += 1
        num_words += len(words)
    return word_counter, word_list, num_lines, num_words


def build_dict(word_counter, vocab_size = 200000):
    count = [['PAD', 0], ['UNK', 1], ['START', 2], ['END', 3]]
    count.extend(word_counter.most_common(vocab_size))
    dictionary = dict()
    for word, _ in count:
        dictionary[word] = len(dictionary)
    return dictionary, {word: idx for idx, word in dictionary.items()}


def train_model(
    train_X,
    train_Y_before,
    train_Y_after,
    epoch = 10,
    batch_size = 16,
    embedding_size = 64,
    maxlen = 50,
    **kwargs
):
    word_counter, _, _, _ = counter_words(train_X)
    dictionary, _ = build_dict(word_counter)
    print(len(dictionary))
    _graph = tf.Graph()
    with _graph.as_default():
        model = Model(
            len(dictionary),
            embedding_size,
            1e-3,
            maxlen,
        )
        sess = tf.InteractiveSession()
        saver = tf.train.Saver(tf.global_variables())
        sess.run(tf.global_variables_initializer())
    saver.save(sess, 'skip-wiki/model.ckpt')

    for e in range(epoch):
        pbar = tqdm(range(0, len(train_X), batch_size), desc = 'minibatch loop')
        for i in pbar:
            batch_x = batch_sequence(
                train_X[i : min(i + batch_size, len(train_X))],
                dictionary,
                maxlen = maxlen,
            )
            batch_y_before = batch_sequence(
                train_Y_before[i : min(i + batch_size, len(train_X))],
                dictionary,
                maxlen = maxlen,
            )
            batch_y_after = batch_sequence(
                train_Y_after[i : min(i + batch_size, len(train_X))],
                dictionary,
                maxlen = maxlen,
            )
            loss, _ = sess.run(
                [model.loss, model.optimizer],
                feed_dict = {
                    model.BEFORE: batch_y_before,
                    model.INPUT: batch_x,
                    model.AFTER: batch_y_after,
                },
            )
            pbar.set_postfix(cost = loss)
    saver.save(sess, 'skip-wiki/model.ckpt')
    return sess, model, dictionary


# In[2]:


def cleaning(string):
    string = re.sub(
        'http\S+|www.\S+',
        '',
        ' '.join(
            [i for i in string.split() if i.find('#') < 0 and i.find('@') < 0]
        ),
    )
    string = unidecode(string).replace('.', '. ').replace(',', ', ')
    string = re.sub('[^A-Za-z ]+', ' ', string)
    string = re.sub(r'[ ]+', ' ', string).strip()
    string = ' '.join(
            [
                i
                for i in re.findall('[\\w\']+|[;:\-\(\)&.,!?"]', string)
                if len(i)
            ]
        )
    return string.lower()

def split_by_dot(string):
    string = re.sub(
        r'(?<!\d)\.(?!\d)',
        'SPLITTT',
        string.replace('\n', '').replace('/', ' '),
    )
    string = string.split('SPLITTT')
    return [re.sub(r'[ ]+', ' ', sentence).strip() for sentence in string]

with open('wiki-ms.txt') as fopen:
    corpus = fopen.read()

print(corpus[:1000])
splitted = corpus.split()
corpus = []
for i in range(0, len(splitted), 50):
    corpus.append(' '.join(splitted[i:i+50]))
print(len(corpus))
corpus = corpus[100000:300000]
corpus = [cleaning(sentence) for sentence in corpus]

print(len(corpus))
corpus = [sentence for sentence in corpus if len(sentence) > 10]
print(len(corpus))

# In[3]:


stride = 1
t_range = int((len(corpus) - 3) / stride + 1)
left, middle, right = [], [], []
for i in range(t_range):
    slices = corpus[i * stride : i * stride + 3]
    left.append(slices[0])
    middle.append(slices[1])
    right.append(slices[2])


# In[5]:


len(left) == len(middle) == len(right)


# In[6]:


from sklearn.utils import shuffle
left, middle, right = shuffle(left, middle, right)


# In[ ]:


_,_,dictionary = train_model(middle,left,right)
with open('skip-wiki-dict.json', 'w') as fopen:
    fopen.write(json.dumps(dictionary))


# coding=utf-8
# Copyright 2018 The Google AI Language Team Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Functions and classes related to optimization (weight updates)."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import re
import tensorflow as tf


def create_optimizer(loss, init_lr, num_train_steps, num_warmup_steps, use_tpu):
    """Creates an optimizer training op."""
    global_step = tf.train.get_or_create_global_step()

    learning_rate = tf.constant(value = init_lr, shape = [], dtype = tf.float32)

    # Implements linear decay of the learning rate.
    learning_rate = tf.train.polynomial_decay(
        learning_rate,
        global_step,
        num_train_steps,
        end_learning_rate = 0.0,
        power = 1.0,
        cycle = False,
    )

    # Implements linear warmup. I.e., if global_step < num_warmup_steps, the
    # learning rate will be `global_step/num_warmup_steps * init_lr`.
    if num_warmup_steps:
        global_steps_int = tf.cast(global_step, tf.int32)
        warmup_steps_int = tf.constant(num_warmup_steps, dtype = tf.int32)

        global_steps_float = tf.cast(global_steps_int, tf.float32)
        warmup_steps_float = tf.cast(warmup_steps_int, tf.float32)

        warmup_percent_done = global_steps_float / warmup_steps_float
        warmup_learning_rate = init_lr * warmup_percent_done

        is_warmup = tf.cast(global_steps_int < warmup_steps_int, tf.float32)
        learning_rate = (
            1.0 - is_warmup
        ) * learning_rate + is_warmup * warmup_learning_rate

    # It is recommended that you use this optimizer for fine tuning, since this
    # is how the model was trained (note that the Adam m/v variables are NOT
    # loaded from init_checkpoint.)
    optimizer = AdamWeightDecayOptimizer(
        learning_rate = learning_rate,
        weight_decay_rate = 0.01,
        beta_1 = 0.9,
        beta_2 = 0.999,  # 0.98 ONLY USED FOR PRETRAIN. MUST CHANGE AT FINE-TUNING 0.999,
        epsilon = 1e-6,
        exclude_from_weight_decay = ['LayerNorm', 'layer_norm', 'bias'],
    )

    if use_tpu:
        optimizer = tf.contrib.tpu.CrossShardOptimizer(optimizer)

    tvars = tf.trainable_variables()
    grads = tf.gradients(loss, tvars)

    # This is how the model was pre-trained.
    (grads, _) = tf.clip_by_global_norm(grads, clip_norm = 1.0)

    train_op = optimizer.apply_gradients(
        zip(grads, tvars), global_step = global_step
    )

    # Normally the global step update is done inside of `apply_gradients`.
    # However, `AdamWeightDecayOptimizer` doesn't do this. But if you use
    # a different optimizer, you should probably take this line out.
    new_global_step = global_step + 1
    train_op = tf.group(train_op, [global_step.assign(new_global_step)])
    return train_op


class AdamWeightDecayOptimizer(tf.train.Optimizer):
    """A basic Adam optimizer that includes "correct" L2 weight decay."""

    def __init__(
        self,
        learning_rate,
        weight_decay_rate = 0.0,
        beta_1 = 0.9,
        beta_2 = 0.999,
        epsilon = 1e-6,
        exclude_from_weight_decay = None,
        name = 'AdamWeightDecayOptimizer',
    ):
        """Constructs a AdamWeightDecayOptimizer."""
        super(AdamWeightDecayOptimizer, self).__init__(False, name)

        self.learning_rate = learning_rate
        self.weight_decay_rate = weight_decay_rate
        self.beta_1 = beta_1
        self.beta_2 = beta_2
        self.epsilon = epsilon
        self.exclude_from_weight_decay = exclude_from_weight_decay

    def apply_gradients(self, grads_and_vars, global_step = None, name = None):
        """See base class."""
        assignments = []
        for (grad, param) in grads_and_vars:
            if grad is None or param is None:
                continue

            param_name = self._get_variable_name(param.name)

            m = tf.get_variable(
                name = param_name + '/adam_m',
                shape = param.shape.as_list(),
                dtype = tf.float32,
                trainable = False,
                initializer = tf.zeros_initializer(),
            )
            v = tf.get_variable(
                name = param_name + '/adam_v',
                shape = param.shape.as_list(),
                dtype = tf.float32,
                trainable = False,
                initializer = tf.zeros_initializer(),
            )

            # Standard Adam update.
            next_m = tf.multiply(self.beta_1, m) + tf.multiply(
                1.0 - self.beta_1, grad
            )
            next_v = tf.multiply(self.beta_2, v) + tf.multiply(
                1.0 - self.beta_2, tf.square(grad)
            )

            update = next_m / (tf.sqrt(next_v) + self.epsilon)

            # Just adding the square of the weights to the loss function is *not*
            # the correct way of using L2 regularization/weight decay with Adam,
            # since that will interact with the m and v parameters in strange ways.
            #
            # Instead we want ot decay the weights in a manner that doesn't interact
            # with the m/v parameters. This is equivalent to adding the square
            # of the weights to the loss with plain (non-momentum) SGD.
            if self._do_use_weight_decay(param_name):
                update += self.weight_decay_rate * param

            update_with_lr = self.learning_rate * update

            next_param = param - update_with_lr

            assignments.extend(
                [param.assign(next_param), m.assign(next_m), v.assign(next_v)]
            )
        return tf.group(*assignments, name = name)

    def _do_use_weight_decay(self, param_name):
        """Whether to use L2 weight decay for `param_name`."""
        if not self.weight_decay_rate:
            return False
        if self.exclude_from_weight_decay:
            for r in self.exclude_from_weight_decay:
                if re.search(r, param_name) is not None:
                    return False
        return True

    def _get_variable_name(self, param_name):
        """Get the variable name from the tensor name."""
        m = re.match('^(.*):\\d+$', param_name)
        if m is not None:
            param_name = m.group(1)
        return param_name



import re
import emoji
from gensim.utils import deaccent
from collections import Counter
from collections import ChainMap
from bs4 import BeautifulSoup
from tqdm import tqdm
import numpy as np
from multiprocessing import Pool
import itertools
from rules import *

re_3986_enhanced = re.compile(
    r"""
        # Parse and capture RFC-3986 Generic URI components.
        ^                                    # anchor to beginning of string
        (?:  (?P<scheme>    [^:/?#\s]+):// )?  # capture optional scheme
        (?:(?P<authority>  [^/?#\s]*)  )?  # capture optional authority
             (?P<path>        [^?#\s]*)      # capture required path
        (?:\?(?P<query>        [^#\s]*)  )?  # capture optional query
        (?:\#(?P<fragment>      [^\s]*)  )?  # capture optional fragment
        $                                    # anchor to end of string
        """,
    re.MULTILINE | re.VERBOSE,
)

re_domain = re.compile(
    r"""
        # Pick out top two levels of DNS domain from authority.
        (?P<domain>[^.]+\.[A-Za-z]{2,6})  # $domain: top two domain levels.
        (?::[0-9]*)?                      # Optional port number.
        $                                 # Anchor to end of string.
        """,
    re.MULTILINE | re.VERBOSE,
)


def domain_search(text):
    try:
        return re_domain.search(
            re_3986_enhanced.match(text).group('authority')
        ).group('domain')
    except:
        return 'url'


def make_cleaning(s, c_dict):
    s = s.translate(c_dict)
    return s


def make_dict_cleaning(s, w_dict):
    s = w_dict.get(s, s)
    return s


def cleaning(string):
    string = ' '.join(
        [make_cleaning(w, normalized_chars) for w in string.split()]
    )
    string = re.sub('\(dot\)', '.', string)
    string = deaccent(string)

    # remove href
    string = (
        re.sub(re.findall(r'\<a(.*?)\>', string)[0], '', string)
        if (len(re.findall(r'\<a (.*?)\>', string)) > 0)
        and ('href' in re.findall(r'\<a (.*?)\>', string)[0])
        else string
    )

    string = re.sub(
        r'\w+:\/{2}[\d\w-]+(\.[\d\w-]+)*(?:(?:\/[^\s/]*))*', '', string
    )
    string = re.sub(r'http\S+|www.\S+', '', string)
    string = re.sub(r'[ ]+', ' ', string).strip().split()
    string = [w for w in string if w[0] != '@']

    return ' '.join(string)


def cleaning_strings(strings):
    for i in tqdm(range(len(strings))):
        strings[i] = cleaning(strings[i])
    return strings


def clean_chars(strings):
    global_chars_list = list(set([c for line in strings for c in line]))
    chars = ''.join(
        [
            c
            for c in global_chars_list
            if (c not in emoji.UNICODE_EMOJI) and (c not in white_list_chars)
        ]
    )
    chars_dict = {}

    for char in chars:
        try:
            new_char = unicodedata.name(char).split()[-1:][0].lower()
            if len(new_char) == 1:
                chars_dict[ord(char)] = new_char
            else:
                chars_dict[ord(char)] = ''
        except:
            chars_dict[ord(char)] = ''

    for i in tqdm(range(len(strings))):
        strings[i] = ' '.join(
            [make_cleaning(w, chars_dict) for w in strings[i].split()]
        )

    return strings


def chunks(l, n):
    for i in range(0, len(l), n):
        yield l[i : i + n]


def chunks_multiple(l, n, dict):
    for i in range(0, len(l), n):
        yield (l[i : i + n], dict)


def multiprocessing(strings, function, cores = 16, list_mode = True):
    df_split = chunks(strings, len(strings) // cores)
    pool = Pool(cores)
    pooled = pool.map(function, df_split)
    pool.close()
    pool.join()
    if list_mode:
        return list(itertools.chain(*pooled))
    else:
        return dict(ChainMap(*pooled))


def multiprocessing_multiple(
    strings, dict, function, cores = 16, list_mode = True
):
    df_split = chunks_multiple(strings, len(strings) // cores, dict)
    pool = Pool(cores)
    pooled = pool.starmap(function, df_split)
    pool.close()
    pool.join()
    if list_mode:
        return list(itertools.chain(*pooled))
    else:
        return dict(ChainMap(*pooled))


def unique_words(strings):
    return list(set([c for line in strings for c in line.split()]))


def duplicate_dots_marks_exclamations(strings):
    temp_dict = {}
    for word in strings:
        new_word = word
        if (
            (Counter(word)['.'] > 1)
            or (Counter(word)['!'] > 1)
            or (Counter(word)['?'] > 1)
            or (Counter(word)[','] > 1)
        ):
            if Counter(word)['.'] > 1:
                new_word = re.sub('\.\.+', ' . . . ', new_word)
            if Counter(word)['!'] > 1:
                new_word = re.sub('\!\!+', ' ! ! ! ', new_word)
            if Counter(word)['?'] > 1:
                new_word = re.sub('\?\?+', ' ? ? ? ', new_word)
            if Counter(word)[','] > 1:
                new_word = re.sub('\,\,+', ' , , , ', new_word)
            temp_dict[word] = new_word
    temp_dict = {k: v for k, v in temp_dict.items() if k != v}
    return temp_dict


def remove_underscore(strings):
    temp_dict = {}
    for word in strings:
        if (
            len(re.compile("[a-zA-Z0-9\-\.\,\/']").sub('', word)) / len(word)
            > 0.6
        ) and ('_' in word):
            temp_dict[word] = re.sub('_', '', word)
    return temp_dict


def isolate_spamchars(strings):
    temp_dict = {}
    for word in strings:
        if (
            (
                len(re.compile("[a-zA-Z0-9\-\.\,\/']").sub('', word))
                / len(word)
                > 0.6
            )
            and (len(Counter(word)) == 1)
            and (len(word) > 2)
        ):
            temp_dict[word] = ' '.join(
                [' ' + next(iter(Counter(word).keys())) + ' ' for i in range(3)]
            )
    return temp_dict


def break_short_words(strings):
    strings = [k for k in strings if len(k) <= 20]
    temp_dict = {}
    for word in strings:
        if '/' in word:
            temp_dict[word] = re.sub('/', ' / ', word)
    return temp_dict


def break_long_words(strings):
    strings = [k for k in strings if len(k) > 20]
    temp_dict = {}
    for word in strings:
        if '_' in word:
            temp_dict[word] = re.sub('_', ' ', word)
        elif '/' in word:
            temp_dict[word] = re.sub('/', ' / ', word)
        elif len(' '.join(word.split('-')).split()) > 2:
            temp_dict[word] = re.sub('-', ' ', word)
    return temp_dict


def remove_ending_underscore(strings):
    strings = [k for k in strings if '_' in k]
    temp_dict = {}
    for word in strings:
        new_word = word
        if word[len(word) - 1] == '_':
            for i in range(len(word), 0, -1):
                if word[i - 1] != '_':
                    new_word = word[:i]
                    temp_dict[word] = new_word
                    break
    return temp_dict


def remove_starting_underscore(strings):
    strings = [k for k in strings if '_' in k]
    temp_dict = {}
    for word in strings:
        new_word = word
        if word[0] == '_':
            for i in range(len(word)):
                if word[i] != '_':
                    new_word = word[i:]
                    temp_dict[word] = new_word
                    break
    return temp_dict


def end_punct(strings):
    strings = [k for k in strings if not k[len(k) - 1].isalnum()]
    temp_dict = {}
    for word in strings:
        new_word = word
        for i in range(len(word), 0, -1):
            if word[i - 1].isalnum():
                new_word = word[:i] + ' ' + word[i:]
                break
        temp_dict[word] = new_word
    temp_dict = {k: v for k, v in temp_dict.items() if k != v}
    return temp_dict


def start_punct(strings):
    strings = [k for k in strings if not k[0].isalnum()]
    temp_dict = {}
    for word in strings:
        new_word = word
        for i in range(len(word)):
            if word[i].isalnum():
                new_word = word[:i] + ' ' + word[i:]
                break
        temp_dict[word] = new_word
    temp_dict = {k: v for k, v in temp_dict.items() if k != v}
    return temp_dict


def join_dashes(strings):
    temp_dict = {}
    for word in strings:
        temp_dict[word] = re.sub('\-\-+', '-', word)
    temp_dict = {k: v for k, v in temp_dict.items() if k != v}
    return temp_dict


def string_dict_cleaning(strings, dict):
    for i in tqdm(range(len(strings))):
        strings[i] = ' '.join(
            [make_dict_cleaning(w, dict) for w in strings[i].split()]
        )
    return strings



import re
import os
import random
import numpy as np
import json
from .texts._jarowrinkler import JaroWinkler
from sklearn.metrics.pairwise import (
    cosine_similarity,
    euclidean_distances,
    manhattan_distances,
)
from ._utils._utils import (
    check_file,
    load_graph,
    check_available,
    generate_session,
    sentencepiece_tokenizer_bert,
    sentencepiece_tokenizer_xlnet,
)
from .preprocessing import _tokenizer
from ._models._bert_model import SIAMESE_BERT
from ._models._xlnet_model import SIAMESE_XLNET
from ._utils._paths import PATH_SIMILARITY, S3_PATH_SIMILARITY


class _VECTORIZER_SIMILARITY:
    def __init__(self, vectorizer):
        self._vectorizer = vectorizer

    def _predict(self, left_strings, right_strings, similarity = 'cosine'):
        if not isinstance(left_strings, list):
            raise ValueError('left_strings must be a list')
        if not isinstance(left_strings[0], str):
            raise ValueError('left_strings must be list of strings')
        if not isinstance(right_strings, list):
            raise ValueError('right_strings must be a list')
        if not isinstance(right_strings[0], str):
            raise ValueError('right_strings must be list of strings')

        if len(left_strings) != len(right_strings):
            raise ValueError(
                'length list of left strings must be same with length list of right strings'
            )
        identical = left_strings == right_strings

        similarity = similarity.lower()
        if similarity == 'cosine':
            similarity_function = cosine_similarity
        elif similarity == 'euclidean':
            similarity_function = euclidean_distances
        elif similarity == 'manhattan':
            similarity_function = manhattan_distances
        else:
            raise ValueError(
                "similarity only supports 'cosine', 'euclidean', and 'manhattan'"
            )

        left_vectors = self._vectorizer.vectorize(left_strings)

        if identical:
            similar = similarity_function(left_vectors, left_vectors)
        else:
            right_vectors = self._vectorizer.vectorize(right_strings)
            similar = similarity_function(left_vectors, right_vectors)

        if similarity == 'cosine':
            return (similar + 1) / 2
        else:
            return 1 / (similar + 1)

    def predict(self, left_string, right_string, similarity = 'cosine'):
        """
        calculate similarity for two different texts.

        Parameters
        ----------
        left_string : str
        right_string : str
        similarity : str, optional (default='mean')
            similarity supported. Allowed values:

            * ``'cosine'`` - cosine similarity.
            * ``'euclidean'`` - euclidean similarity.
            * ``'manhattan'`` - manhattan similarity.

        Returns
        -------
        float: float
        """
        if not isinstance(left_string, str):
            raise ValueError('left_string must be a string')
        if not isinstance(right_string, str):
            raise ValueError('right_string must be a string')
        return self._predict(
            [left_string], [right_string], similarity = similarity
        )[0, 0]

    def predict_batch(self, left_strings, right_strings, similarity = 'cosine'):
        """
        calculate similarity for two different batch of texts.

        Parameters
        ----------
        left_strings : list of str
        right_strings : list of str
        similarity : str, optional (default='mean')
            similarity supported. Allowed values:

            * ``'cosine'`` - cosine similarity.
            * ``'euclidean'`` - euclidean similarity.
            * ``'manhattan'`` - manhattan similarity.

        Returns
        -------
        list: list of float
        """
        return self._predict(
            left_strings, right_strings, similarity = similarity
        ).diagonal()

    def tree_plot(
        self,
        strings,
        similarity = 'cosine',
        visualize = True,
        figsize = (7, 7),
        annotate = True,
    ):
        """
        plot a tree plot based on output from bert similarity.

        Parameters
        ----------
        strings : list of str
            list of strings.
        similarity : str, optional (default='mean')
            similarity supported. Allowed values:

            * ``'cosine'`` - cosine similarity.
            * ``'euclidean'`` - euclidean similarity.
            * ``'manhattan'`` - manhattan similarity.
        visualize : bool
            if True, it will render plt.show, else return data.
        figsize : tuple, (default=(7, 7))
            figure size for plot.

        Returns
        -------
        list_dictionaries: list of results
        """
        if not isinstance(visualize, bool):
            raise ValueError('visualize must be a boolean')
        if not isinstance(figsize, tuple):
            raise ValueError('figsize must be a tuple')
        if not isinstance(annotate, bool):
            raise ValueError('annotate must be a boolean')

        results = self._predict(strings, strings, similarity = similarity)
        if not visualize:
            return results

        try:
            import matplotlib.pyplot as plt
            import seaborn as sns

            sns.set()
        except:
            raise Exception(
                'matplotlib and seaborn not installed. Please install it and try again.'
            )

        plt.figure(figsize = figsize)
        g = sns.clustermap(
            results,
            cmap = 'Blues',
            xticklabels = strings,
            yticklabels = strings,
            annot = annotate,
        )
        plt.show()


class _DOC2VEC_SIMILARITY:
    def __init__(self, vectorizer):
        self._vectorizer = vectorizer
        self._jarowinkler = JaroWinkler()

    def _predict(
        self,
        left_strings,
        right_strings,
        aggregation = 'mean',
        similarity = 'cosine',
        soft = True,
    ):

        if not isinstance(left_strings, list):
            raise ValueError('left_strings must be a list')
        if not isinstance(left_strings[0], str):
            raise ValueError('left_strings must be list of strings')
        if not isinstance(right_strings, list):
            raise ValueError('right_strings must be a list')
        if not isinstance(right_strings[0], str):
            raise ValueError('right_strings must be list of strings')
        if not isinstance(aggregation, str):
            raise ValueError('aggregation must be a string')
        if not isinstance(similarity, str):
            raise ValueError('similarity must be a string')
        if not isinstance(soft, bool):
            raise ValueError('soft must be a boolean')

        if len(left_strings) != len(right_strings):
            raise ValueError(
                'length list of left strings must be same with length list of right strings'
            )
        identical = left_strings == right_strings

        aggregation = aggregation.lower()
        if aggregation == 'mean':
            aggregation_function = np.mean
        elif aggregation == 'min':
            aggregation_function = np.min
        elif aggregation == 'max':
            aggregation_function = np.max
        elif aggregation == 'sum':
            aggregation_function = np.sum
        elif aggregation == 'sqrt':
            aggregation_function = np.sqrt
        else:
            raise ValueError(
                "aggregation only supports 'mean', 'min', 'max', 'sum' and 'sqrt'"
            )

        similarity = similarity.lower()
        if similarity == 'cosine':
            similarity_function = cosine_similarity
        elif similarity == 'euclidean':
            similarity_function = euclidean_distances
        elif similarity == 'manhattan':
            similarity_function = manhattan_distances
        else:
            raise ValueError(
                "similarity only supports 'cosine', 'euclidean', and 'manhattan'"
            )

        left_vectors, right_vectors = [], []
        for i in range(len(left_strings)):
            left_string = left_strings[i]
            right_string = right_strings[i]
            left_tokenized = _tokenizer(left_string)
            if not len(left_tokenized):
                raise ValueError('insert not empty left string')
            right_tokenized = _tokenizer(right_string)
            if not len(right_tokenized):
                raise ValueError('insert not empty right string')

            in_vector = []
            for token in left_tokenized:
                try:
                    in_vector.append(self._vectorizer.get_vector_by_name(token))
                except:
                    if not soft:
                        pass
                    else:
                        arr = np.array(
                            [
                                self._jarowinkler.similarity(token, k)
                                for k in self._vectorizer.words
                            ]
                        )
                        idx = (-arr).argsort()[0]
                        in_vector.append(
                            self._vectorizer.get_vector_by_name(
                                self._vectorizer.words[idx]
                            )
                        )
            left_vectors.append(aggregation_function(in_vector, axis = 0))

            if not identical:
                in_vector = []
                for token in right_tokenized:
                    try:
                        in_vector.append(
                            self._vectorizer.get_vector_by_name(token)
                        )
                    except:
                        if not soft:
                            pass
                        else:
                            arr = np.array(
                                [
                                    self._jarowinkler.similarity(token, k)
                                    for k in self._vectorizer.words
                                ]
                            )
                            idx = (-arr).argsort()[0]
                            in_vector.append(
                                self._vectorizer.get_vector_by_name(
                                    self._vectorizer.words[idx]
                                )
                            )

                right_vectors.append(aggregation_function(in_vector, axis = 0))

        if identical:
            similar = similarity_function(left_vectors, left_vectors)
        else:
            similar = similarity_function(left_vectors, right_vectors)

        if similarity == 'cosine':
            return (similar + 1) / 2
        else:
            return 1 / (similar + 1)

    def predict(
        self,
        left_string,
        right_string,
        aggregation = 'mean',
        similarity = 'cosine',
        soft = True,
    ):
        """
        calculate similarity for two different texts.

        Parameters
        ----------
        left_string : str
        right_string : str
        aggregation : str, optional (default='mean')
            Aggregation supported. Allowed values:

            * ``'mean'`` - mean.
            * ``'min'`` - min.
            * ``'max'`` - max.
            * ``'sum'`` - sum.
            * ``'sqrt'`` - square root.
        similarity : str, optional (default='mean')
            similarity supported. Allowed values:

            * ``'cosine'`` - cosine similarity.
            * ``'euclidean'`` - euclidean similarity.
            * ``'manhattan'`` - manhattan similarity.
        soft: bool, optional (default=True)
            word not inside word vector will replace with nearest word if True, else, will skip.

        Returns
        -------
        float: float
        """
        if not isinstance(left_string, str):
            raise ValueError('left_string must be a string')
        if not isinstance(right_string, str):
            raise ValueError('right_string must be a string')

        return self._predict(
            [left_string],
            [right_string],
            aggregation = aggregation,
            similarity = similarity,
            soft = soft,
        )[0, 0]

    def predict_batch(
        self,
        left_strings,
        right_strings,
        aggregation = 'mean',
        similarity = 'cosine',
        soft = True,
    ):
        """
        calculate similarity for two different batch of texts.

        Parameters
        ----------
        left_strings : list of str
        right_strings : list of str
        aggregation : str, optional (default='mean')
            Aggregation supported. Allowed values:

            * ``'mean'`` - mean.
            * ``'min'`` - min.
            * ``'max'`` - max.
            * ``'sum'`` - sum.
            * ``'sqrt'`` - square root.
        similarity : str, optional (default='mean')
            similarity supported. Allowed values:

            * ``'cosine'`` - cosine similarity.
            * ``'euclidean'`` - euclidean similarity.
            * ``'manhattan'`` - manhattan similarity.
        soft: bool, optional (default=True)
            word not inside word vector will replace with nearest word if True, else, will skip.

        Returns
        -------
        list: list of float
        """

        return self._predict(
            left_strings,
            right_strings,
            aggregation = aggregation,
            similarity = similarity,
            soft = soft,
        ).diagonal()

    def tree_plot(
        self,
        strings,
        aggregation = 'mean',
        similarity = 'cosine',
        soft = True,
        visualize = True,
        figsize = (7, 7),
        annotate = True,
    ):
        """
        plot a tree plot based on output from bert similarity.

        Parameters
        ----------
        strings : list of str
            list of strings
        aggregation : str, optional (default='mean')
            Aggregation supported. Allowed values:

            * ``'mean'`` - mean.
            * ``'min'`` - min.
            * ``'max'`` - max.
            * ``'sum'`` - sum.
            * ``'sqrt'`` - square root.
        similarity : str, optional (default='mean')
            similarity supported. Allowed values:

            * ``'cosine'`` - cosine similarity.
            * ``'euclidean'`` - euclidean similarity.
            * ``'manhattan'`` - manhattan similarity.
        soft: bool, optional (default=True)
            word not inside word vector will replace with nearest word if True, else, will skip.
        visualize : bool
            if True, it will render plt.show, else return data.
        figsize : tuple, (default=(7, 7))
            figure size for plot.

        Returns
        -------
        list_dictionaries: list of results
        """
        if not isinstance(visualize, bool):
            raise ValueError('visualize must be a boolean')
        if not isinstance(figsize, tuple):
            raise ValueError('figsize must be a tuple')
        if not isinstance(annotate, bool):
            raise ValueError('annotate must be a boolean')

        results = self._predict(
            strings,
            strings,
            aggregation = aggregation,
            similarity = similarity,
            soft = soft,
        )
        if not visualize:
            return results

        try:
            import matplotlib.pyplot as plt
            import seaborn as sns

            sns.set()
        except:
            raise Exception(
                'matplotlib and seaborn not installed. Please install it and try again.'
            )

        plt.figure(figsize = figsize)
        g = sns.clustermap(
            results,
            cmap = 'Blues',
            xticklabels = strings,
            yticklabels = strings,
            annot = annotate,
        )
        plt.show()


def doc2vec(vectorizer):
    """
    Doc2vec interface for text similarity.

    Parameters
    ----------
    vectorizer : object
        word vector interface object, fast-text, word2vec, elmo.

    Returns
    -------
    _DOC2VEC_SIMILARITY: malaya.similarity._DOC2VEC_SIMILARITY
    """

    if not hasattr(vectorizer, 'get_vector_by_name'):
        raise ValueError('vectorizer must has `get_vector_by_name` method')
    return _DOC2VEC_SIMILARITY(vectorizer)


def encoder(vectorizer):
    """
    Encoder interface for text similarity.

    Parameters
    ----------
    vectorizer : object
        encoder interface object, BERT, skip-thought, XLNET.

    Returns
    -------
    _DOC2VEC_SIMILARITY: malaya.similarity._DOC2VEC_SIMILARITY
    """

    if not hasattr(vectorizer, 'vectorize'):
        raise ValueError('vectorizer must has `vectorize` method')
    return _VECTORIZER_SIMILARITY(vectorizer)


_availability = {'bert': ['base'], 'xlnet': ['base'], 'albert': ['base']}


def available_transformer_model():
    """
    List available transformer similarity models.
    """
    return _availability


def transformer(model = 'bert', size = 'base', validate = True):
    """
    Load Transformer sentiment model.

    Parameters
    ----------
    model : str, optional (default='bert')
        Model architecture supported. Allowed values:

        * ``'bert'`` - BERT architecture from google.
        * ``'xlnet'`` - XLNET architecture from google.
        * ``'albert'`` - ALBERT architecture from google.
    size : str, optional (default='base')
        Model size supported. Allowed values:

        * ``'base'`` - BASE size.
        * ``'small'`` - SMALL size.
    validate: bool, optional (default=True)
        if True, malaya will check model availability and download if not available.

    Returns
    -------
    BERT : malaya._models._bert_model.BINARY_BERT class
    """
    if not isinstance(model, str):
        raise ValueError('model must be a string')
    if not isinstance(size, str):
        raise ValueError('size must be a string')
    if not isinstance(validate, bool):
        raise ValueError('validate must be a boolean')

    model = model.lower()
    size = size.lower()
    if model not in _availability:
        raise Exception(
            'model not supported, please check supported models from malaya.sentiment.available_transformer_model()'
        )
    if size not in _availability[model]:
        raise Exception(
            'size not supported, please check supported models from malaya.sentiment.available_transformer_model()'
        )

    if validate:
        check_file(
            PATH_SIMILARITY[model][size], S3_PATH_SIMILARITY[model][size]
        )
    else:
        if not check_available(PATH_SIMILARITY[model][size]):
            raise Exception(
                'similarity/%s/%s is not available, please `validate = True`'
                % (model, size)
            )

    try:
        g = load_graph(PATH_SIMILARITY[model][size]['model'])
    except:
        raise Exception(
            "model corrupted due to some reasons, please run malaya.clear_cache('similarity/%s/%s') and try again"
            % (model, size)
        )

    if model in ['albert', 'bert']:
        if model == 'bert':
            from ._transformer._bert import _extract_attention_weights_import
        if model == 'albert':
            from ._transformer._albert import _extract_attention_weights_import

        tokenizer, cls, sep = sentencepiece_tokenizer_bert(
            PATH_SIMILARITY[model][size]['tokenizer'],
            PATH_SIMILARITY[model][size]['vocab'],
        )

        return SIAMESE_BERT(
            X = g.get_tensor_by_name('import/Placeholder:0'),
            segment_ids = g.get_tensor_by_name('import/Placeholder_1:0'),
            input_masks = g.get_tensor_by_name('import/Placeholder_2:0'),
            logits = g.get_tensor_by_name('import/logits:0'),
            sess = generate_session(graph = g),
            tokenizer = tokenizer,
            label = ['not similar', 'similar'],
            cls = cls,
            sep = sep,
        )

    if model in ['xlnet']:
        from ._transformer._xlnet import _extract_attention_weights_import

        tokenizer = sentencepiece_tokenizer_xlnet(
            PATH_SIMILARITY[model][size]['tokenizer']
        )

        return SIAMESE_XLNET(
            X = g.get_tensor_by_name('import/Placeholder:0'),
            segment_ids = g.get_tensor_by_name('import/Placeholder_1:0'),
            input_masks = g.get_tensor_by_name('import/Placeholder_2:0'),
            logits = g.get_tensor_by_name('import/logits:0'),
            sess = generate_session(graph = g),
            tokenizer = tokenizer,
            label = ['not similar', 'similar'],
        )



import json
from ._utils._utils import (
    check_file,
    load_graph,
    generate_session,
    check_available,
    sentencepiece_tokenizer_bert,
    sentencepiece_tokenizer_xlnet,
)
from ._utils._parse_dependency import DependencyGraph
from ._utils._paths import PATH_DEPEND, S3_PATH_DEPEND

_dependency_tags = {
    'PAD': 0,
    'X': 1,
    'nsubj': 2,
    'cop': 3,
    'det': 4,
    'root': 5,
    'nsubj:pass': 6,
    'acl': 7,
    'case': 8,
    'obl': 9,
    'flat': 10,
    'punct': 11,
    'appos': 12,
    'amod': 13,
    'compound': 14,
    'advmod': 15,
    'cc': 16,
    'obj': 17,
    'conj': 18,
    'mark': 19,
    'advcl': 20,
    'nmod': 21,
    'nummod': 22,
    'dep': 23,
    'xcomp': 24,
    'ccomp': 25,
    'parataxis': 26,
    'compound:plur': 27,
    'fixed': 28,
    'aux': 29,
    'csubj': 30,
    'iobj': 31,
    'csubj:pass': 32,
}


def dependency_graph(tagging, indexing):
    """
    Return helper object for dependency parser results. Only accept tagging and indexing outputs from dependency models
    """
    result = []
    for i in range(len(tagging)):
        result.append(
            '%d\t%s\t_\t_\t_\t_\t%d\t%s\t_\t_'
            % (i + 1, tagging[i][0], int(indexing[i][1]), tagging[i][1])
        )
    return DependencyGraph('\n'.join(result), top_relation_label = 'root')


_availability = {'bert': ['base'], 'xlnet': ['base'], 'albert': ['base']}


def available_transformer_model():
    """
    List available transformer dependency parsing models.
    """
    return _availability


def transformer(model = 'xlnet', size = 'base', validate = True):
    """
    Load Transformer Entity Tagging model, transfer learning Transformer + biaffine attention.

    Parameters
    ----------
    model : str, optional (default='bert')
        Model architecture supported. Allowed values:

        * ``'bert'`` - BERT architecture from google.
        * ``'xlnet'`` - XLNET architecture from google.
        * ``'albert'`` - ALBERT architecture from google.
    size : str, optional (default='base')
        Model size supported. Allowed values:

        * ``'base'`` - BASE size.
        * ``'small'`` - SMALL size.
    validate: bool, optional (default=True)
        if True, malaya will check model availability and download if not available.

    Returns
    -------
    MODEL : Transformer class
    """
    if not isinstance(model, str):
        raise ValueError('model must be a string')
    if not isinstance(size, str):
        raise ValueError('size must be a string')
    if not isinstance(validate, bool):
        raise ValueError('validate must be a boolean')

    model = model.lower()
    size = size.lower()
    if model not in _availability:
        raise Exception(
            'model not supported, please check supported models from malaya.dependency.available_transformer_model()'
        )
    if size not in _availability[model]:
        raise Exception(
            'size not supported, please check supported models from malaya.dependency.available_transformer_model()'
        )

    if validate:
        check_file(PATH_DEPEND[model][size], S3_PATH_DEPEND[model][size])
    else:
        if not check_available(PATH_DEPEND[model][size]):
            raise Exception(
                'dependency/%s/%s is not available, please `validate = True`'
                % (model, size)
            )

    try:
        g = load_graph(PATH_DEPEND[model][size]['model'])
    except:
        raise Exception(
            "model corrupted due to some reasons, please run malaya.clear_cache('dependency/%s/%s') and try again"
            % (model, size)
        )

    if model in ['bert', 'albert']:
        from ._models._bert_model import DEPENDENCY_BERT

        tokenizer, cls, sep = sentencepiece_tokenizer_bert(
            PATH_DEPEND[model][size]['tokenizer'],
            PATH_DEPEND[model][size]['vocab'],
        )

        return DEPENDENCY_BERT(
            X = g.get_tensor_by_name('import/Placeholder:0'),
            segment_ids = None,
            input_masks = None,
            logits = g.get_tensor_by_name('import/logits:0'),
            sess = generate_session(graph = g),
            tokenizer = tokenizer,
            cls = cls,
            sep = sep,
            settings = _dependency_tags,
            heads_seq = g.get_tensor_by_name('import/heads_seq:0'),
        )

    if model in ['xlnet']:
        from ._models._xlnet_model import DEPENDENCY_XLNET

        tokenizer = sentencepiece_tokenizer_xlnet(
            PATH_DEPEND[model][size]['tokenizer']
        )

        return DEPENDENCY_XLNET(
            X = g.get_tensor_by_name('import/Placeholder:0'),
            segment_ids = g.get_tensor_by_name('import/Placeholder_1:0'),
            input_masks = g.get_tensor_by_name('import/Placeholder_2:0'),
            logits = g.get_tensor_by_name('import/logits:0'),
            sess = generate_session(graph = g),
            tokenizer = tokenizer,
            settings = _dependency_tags,
            heads_seq = g.get_tensor_by_name('import/heads_seq:0'),
        )



from flask import Flask, request, jsonify
import malaya

app = Flask(__name__)

model = malaya.sentiment.transformer(
    model = 'albert', size = 'base', validate = False
)


@app.route('/', methods = ['GET'])
def index():
    strings = [request.args.get('string')] * 50
    r = model.predict_batch(strings, get_proba = True)
    return jsonify('done')


application = app



import malaya
string = 'KUALA LUMPUR: Sempena sambutan Aidilfitri minggu depan'
malaya.get_available_entities_models()

def test_multinomial_entities():
    results = malaya.multinomial_entities(string)
    assert len(results)

def test_xgb_entities():
    results = malaya.xgb_entities(string)
    assert len(results)

def test_deep_entities_char():
    model = malaya.deep_entities('char')
    assert len(model.predict(string))

def test_deep_entities_word():
    model = malaya.deep_entities('word')
    assert len(model.predict(string))

def test_deep_entities_concat():
    model = malaya.deep_entities('concat')
    assert len(model.predict(string))



from __future__ import print_function, unicode_literals

malaysian_number_system = {
    'kosong': 0,
    'satu': 1,
    'dua': 2,
    'tiga': 3,
    'empat': 4,
    'lima': 5,
    'enam': 6,
    'tujuh': 7,
    'lapan': 8,
    'sembilan': 9,
    'sepuluh': 10,
    'seribu': 1000,
    'sejuta': 1000000,
    'seratus': 100,
    'sebelas': 11,
    'ratus': 100,
    'ribu': 1000,
    'juta': 1000000,
    'bilion': 1000000000,
    'perpuluhan': '.',
    'negatif': -1,
    'belas': 10,
    'puluh': 10,
    'pertama': 1,
}

decimal_words = [
    'kosong',
    'satu',
    'dua',
    'tiga',
    'empat',
    'lima',
    'enam',
    'tujuh',
    'lapan',
    'sembilan',
]


def _get_decimal_sum(decimal_digit_words):
    decimal_number_str = []
    for dec_word in decimal_digit_words:
        if dec_word not in decimal_words:
            return 0
        else:
            decimal_number_str.append(malaysian_number_system[dec_word])
    final_decimal_string = '0.' + ''.join(map(str, decimal_number_str))
    return float(final_decimal_string)


def _number_formation(number_words):
    numbers = []
    belas = False
    for number_word in number_words:
        if number_word in ['belas', 'sebelas']:
            belas = True
        numbers.append(malaysian_number_system[number_word])
    if len(numbers) == 5:
        return (
            (numbers[0] * numbers[1]) + (numbers[2] * numbers[3]) + numbers[4]
        )
    elif len(numbers) == 4:
        if numbers[0] == 100:
            return numbers[0] + (numbers[1] * numbers[2]) + numbers[3]
        return (numbers[0] * numbers[1]) + numbers[2] + numbers[3]
    elif len(numbers) == 3:
        if belas:
            return numbers[0] + numbers[1] + numbers[2]
        return numbers[0] * numbers[1] + numbers[2]
    elif len(numbers) == 2:
        if 100 in numbers or 10 in numbers:
            if belas:
                return numbers[0] + numbers[1]
            return numbers[0] * numbers[1]
        else:
            return numbers[0] + numbers[1]
    else:
        return numbers[0]


def word2num(string):
    if not isinstance(string, str):
        raise ValueError('input must be a string')

    string = string.replace('-', ' ')
    string = string.replace('ke', '')
    string = string.replace('dan', '')
    string = string.lower()

    if string.isdigit():
        return int(string)

    split_words = string.strip().split()

    clean_numbers = []
    clean_decimal_numbers = []

    for word in split_words:
        if word in malaysian_number_system:
            clean_numbers.append(word)

    if not len(clean_numbers):
        raise ValueError(
            'No valid number words found! Please enter a valid number word'
        )

    if (
        clean_numbers.count('ribu') > 1
        or clean_numbers.count('juta') > 1
        or clean_numbers.count('bilion') > 1
        or clean_numbers.count('perpuluhan') > 1
        or clean_numbers.count('negatif') > 1
        or clean_numbers.count('seribu') > 1
        or clean_numbers.count('sejuta') > 1
    ):
        raise ValueError(
            'Redundant number word! Please enter a valid number word'
        )

    negative = False
    if clean_numbers[0] == 'negatif':
        negative = True
        clean_numbers = clean_numbers[1:]

    if clean_numbers.count('perpuluhan') == 1:
        clean_decimal_numbers = clean_numbers[
            clean_numbers.index('perpuluhan') + 1 :
        ]
        clean_numbers = clean_numbers[: clean_numbers.index('perpuluhan')]

    billion_index = (
        clean_numbers.index('bilion') if 'bilion' in clean_numbers else -1
    )
    million_index = (
        clean_numbers.index('juta') if 'juta' in clean_numbers else -1
    )
    thousand_index = (
        clean_numbers.index('ribu') if 'ribu' in clean_numbers else -1
    )

    if (
        thousand_index > -1
        and (thousand_index < million_index or thousand_index < billion_index)
    ) or (million_index > -1 and million_index < billion_index):
        raise ValueError('Malformed number! Please enter a valid number word')

    total_sum = 0

    if len(clean_numbers) > 0:
        if len(clean_numbers) == 1:
            total_sum += malaysian_number_system[clean_numbers[0]]
        else:
            if billion_index > -1:
                billion_multiplier = _number_formation(
                    clean_numbers[0:billion_index]
                )
                total_sum += billion_multiplier * 1000000000

            if million_index > -1:
                if billion_index > -1:
                    million_multiplier = _number_formation(
                        clean_numbers[billion_index + 1 : million_index]
                    )
                else:
                    million_multiplier = _number_formation(
                        clean_numbers[0:million_index]
                    )
                total_sum += million_multiplier * 1000000

            if thousand_index > -1:
                if million_index > -1:
                    thousand_multiplier = _number_formation(
                        clean_numbers[million_index + 1 : thousand_index]
                    )
                elif billion_index > -1 and million_index == -1:
                    thousand_multiplier = _number_formation(
                        clean_numbers[billion_index + 1 : thousand_index]
                    )
                else:
                    thousand_multiplier = _number_formation(
                        clean_numbers[0:thousand_index]
                    )
                total_sum += thousand_multiplier * 1000

            if thousand_index > -1 and thousand_index != len(clean_numbers) - 1:
                hundreds = _number_formation(
                    clean_numbers[thousand_index + 1 :]
                )
            elif million_index > -1 and million_index != len(clean_numbers) - 1:
                hundreds = _number_formation(clean_numbers[million_index + 1 :])
            elif billion_index > -1 and billion_index != len(clean_numbers) - 1:
                hundreds = _number_formation(clean_numbers[billion_index + 1 :])
            elif (
                thousand_index == -1
                and million_index == -1
                and billion_index == -1
            ):
                hundreds = _number_formation(clean_numbers)
            else:
                hundreds = 0
            total_sum += hundreds

    if len(clean_decimal_numbers) > 0:
        decimal_sum = _get_decimal_sum(clean_decimal_numbers)
        total_sum += decimal_sum

    return total_sum * -1 if negative else total_sum



import re
from .texts._tatabahasa import tatabahasa_dict, hujung, permulaan
from ._utils import _tag_class
from ._utils._paths import PATH_POS, S3_PATH_POS


_availability = {
    'bert': ['base', 'small'],
    'xlnet': ['base'],
    'albert': ['base'],
}


def available_transformer_model():
    """
    List available transformer Part-Of-Speech Tagging models.
    """
    return _availability


def _naive_POS_word(word):
    for key, vals in tatabahasa_dict.items():
        if word in vals:
            return (key, word)
    try:
        if len(re.findall(r'^(.*?)(%s)$' % ('|'.join(hujung[:1])), i)[0]) > 1:
            return ('KJ', word)
    except:
        pass
    try:
        if (
            len(re.findall(r'^(.*?)(%s)' % ('|'.join(permulaan[:-4])), word)[0])
            > 1
        ):
            return ('KJ', word)
    except Exception as e:
        pass
    if len(word) > 2:
        return ('KN', word)
    else:
        return ('', word)


def naive(string):
    """
    Recognize POS in a string using Regex.

    Parameters
    ----------
    string: str

    Returns
    -------
    string : tokenized string with POS related
    """
    if not isinstance(string, str):
        raise ValueError('input must be a string')
    string = string.lower()
    results = []
    for i in string.split():
        results.append(_naive_POS_word(i))
    return results


def transformer(model = 'xlnet', size = 'base', validate = True):
    """
    Load Transformer POS Tagging model, transfer learning Transformer + CRF.

    Parameters
    ----------
    model : str, optional (default='bert')
        Model architecture supported. Allowed values:

        * ``'bert'`` - BERT architecture from google.
        * ``'xlnet'`` - XLNET architecture from google.
        * ``'albert'`` - ALBERT architecture from google.
    size : str, optional (default='base')
        Model size supported. Allowed values:

        * ``'base'`` - BASE size.
        * ``'small'`` - SMALL size.
    validate: bool, optional (default=True)
        if True, malaya will check model availability and download if not available.

    Returns
    -------
    MODEL : Transformer class
    """
    if not isinstance(model, str):
        raise ValueError('model must be a string')
    if not isinstance(size, str):
        raise ValueError('size must be a string')
    if not isinstance(validate, bool):
        raise ValueError('validate must be a boolean')

    model = model.lower()
    size = size.lower()
    if model not in _availability:
        raise Exception(
            'model not supported, please check supported models from malaya.pos.available_transformer_model()'
        )
    if size not in _availability[model]:
        raise Exception(
            'size not supported, please check supported models from malaya.pos.available_transformer_model()'
        )
    return _tag_class.transformer(
        PATH_POS,
        S3_PATH_POS,
        'pos',
        model = model,
        size = size,
        validate = validate,
    )



# coding=utf-8
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import unicodedata
import six
from functools import partial


SPIECE_UNDERLINE = '▁'


def printable_text(text):
  """Returns text encoded in a way suitable for print or `tf.logging`."""

  # These functions want `str` for both Python2 and Python3, but in one case
  # it's a Unicode string and in the other it's a byte string.
  if six.PY3:
    if isinstance(text, str):
      return text
    elif isinstance(text, bytes):
      return text.decode("utf-8", "ignore")
    else:
      raise ValueError("Unsupported string type: %s" % (type(text)))
  elif six.PY2:
    if isinstance(text, str):
      return text
    elif isinstance(text, unicode):
      return text.encode("utf-8")
    else:
      raise ValueError("Unsupported string type: %s" % (type(text)))
  else:
    raise ValueError("Not running on Python2 or Python 3?")


def print_(*args):
  new_args = []
  for arg in args:
    if isinstance(arg, list):
      s = [printable_text(i) for i in arg]
      s = ' '.join(s)
      new_args.append(s)
    else:
      new_args.append(printable_text(arg))
  print(*new_args)


def preprocess_text(inputs, lower=False, remove_space=True, keep_accents=False):
  if remove_space:
    outputs = ' '.join(inputs.strip().split())
  else:
    outputs = inputs
  outputs = outputs.replace("``", '"').replace("''", '"')

  if six.PY2 and isinstance(outputs, str):
    outputs = outputs.decode('utf-8')

  if not keep_accents:
    outputs = unicodedata.normalize('NFKD', outputs)
    outputs = ''.join([c for c in outputs if not unicodedata.combining(c)])
  if lower:
    outputs = outputs.lower()

  return outputs


def encode_pieces(sp_model, text, return_unicode=True, sample=False):
  # return_unicode is used only for py2

  # note(zhiliny): in some systems, sentencepiece only accepts str for py2
  if six.PY2 and isinstance(text, unicode):
    text = text.encode('utf-8')

  if not sample:
    pieces = sp_model.EncodeAsPieces(text)
  else:
    pieces = sp_model.SampleEncodeAsPieces(text, 64, 0.1)
  new_pieces = []
  for piece in pieces:
    if len(piece) > 1 and piece[-1] == ',' and piece[-2].isdigit():
      cur_pieces = sp_model.EncodeAsPieces(
          piece[:-1].replace(SPIECE_UNDERLINE, ''))
      if piece[0] != SPIECE_UNDERLINE and cur_pieces[0][0] == SPIECE_UNDERLINE:
        if len(cur_pieces[0]) == 1:
          cur_pieces = cur_pieces[1:]
        else:
          cur_pieces[0] = cur_pieces[0][1:]
      cur_pieces.append(piece[-1])
      new_pieces.extend(cur_pieces)
    else:
      new_pieces.append(piece)

  # note(zhiliny): convert back to unicode for py2
  if six.PY2 and return_unicode:
    ret_pieces = []
    for piece in new_pieces:
      if isinstance(piece, str):
        piece = piece.decode('utf-8')
      ret_pieces.append(piece)
    new_pieces = ret_pieces

  return new_pieces


def encode_ids(sp_model, text, sample=False):
  pieces = encode_pieces(sp_model, text, return_unicode=False, sample=sample)
  ids = [sp_model.PieceToId(piece) for piece in pieces]
  return ids


if __name__ == '__main__':
  import sentencepiece as spm

  sp = spm.SentencePieceProcessor()
  sp.load('sp10m.uncased.v3.model')

  print_(u'I was born in 2000, and this is falsé.')
  print_(u'ORIGINAL', sp.EncodeAsPieces(u'I was born in 2000, and this is falsé.'))
  print_(u'OURS', encode_pieces(sp, u'I was born in 2000, and this is falsé.'))
  print(encode_ids(sp, u'I was born in 2000, and this is falsé.'))
  print_('')
  prepro_func = partial(preprocess_text, lower=True)
  print_(prepro_func('I was born in 2000, and this is falsé.'))
  print_('ORIGINAL', sp.EncodeAsPieces(prepro_func('I was born in 2000, and this is falsé.')))
  print_('OURS', encode_pieces(sp, prepro_func('I was born in 2000, and this is falsé.')))
  print(encode_ids(sp, prepro_func('I was born in 2000, and this is falsé.')))
  print_('')
  print_('I was born in 2000, and this is falsé.')
  print_('ORIGINAL', sp.EncodeAsPieces('I was born in 2000, and this is falsé.'))
  print_('OURS', encode_pieces(sp, 'I was born in 2000, and this is falsé.'))
  print(encode_ids(sp, 'I was born in 2000, and this is falsé.'))
  print_('')
  print_('I was born in 92000, and this is falsé.')
  print_('ORIGINAL', sp.EncodeAsPieces('I was born in 92000, and this is falsé.'))
  print_('OURS', encode_pieces(sp, 'I was born in 92000, and this is falsé.'))
  print(encode_ids(sp, 'I was born in 92000, and this is falsé.'))




import json
import os
import pickle
from ._utils import (
    check_file,
    load_graph,
    check_available,
    generate_session,
    sentencepiece_tokenizer_bert,
    sentencepiece_tokenizer_xlnet,
)
from .._models._sklearn_model import BINARY_BAYES, MULTICLASS_BAYES
from .._models._bert_model import MULTICLASS_BERT, BINARY_BERT
from .._models._xlnet_model import MULTICLASS_XLNET, BINARY_XLNET
from .._transformer._bert import bert_num_layers


def multinomial(path, s3_path, class_name, label, validate = True):
    if validate:
        check_file(path['multinomial'], s3_path['multinomial'])
    else:
        if not check_available(path['multinomial']):
            raise Exception(
                '%s/multinomial is not available, please `validate = True`'
                % (class_name)
            )
    try:
        with open(path['multinomial']['model'], 'rb') as fopen:
            multinomial = pickle.load(fopen)
        with open(path['multinomial']['vector'], 'rb') as fopen:
            vectorize = pickle.load(fopen)
    except:
        raise Exception(
            "model corrupted due to some reasons, please run malaya.clear_cache('%s/multinomial') and try again"
            % (class_name)
        )
    from ..stem import _classification_textcleaning_stemmer

    if len(label) > 2:
        selected_class = MULTICLASS_BAYES
    else:
        selected_class = BINARY_BAYES
    return selected_class(
        multinomial = multinomial,
        label = label,
        vectorize = vectorize,
        cleaning = _classification_textcleaning_stemmer,
    )


def transformer(
    path,
    s3_path,
    class_name,
    label,
    model = 'bert',
    size = 'base',
    validate = True,
):
    if validate:
        check_file(path[model][size], s3_path[model][size])
    else:
        if not check_available(path[model][size]):
            raise Exception(
                '%s/%s/%s is not available, please `validate = True`'
                % (class_name, model, size)
            )

    try:
        g = load_graph(path[model][size]['model'])
    except:
        raise Exception(
            "model corrupted due to some reasons, please run malaya.clear_cache('%s/%s/%s') and try again"
            % (class_name, model, size)
        )

    if len(label) > 2 or class_name == 'relevancy':
        if model in ['albert', 'bert']:
            selected_class = MULTICLASS_BERT
        if model in ['xlnet']:
            selected_class = MULTICLASS_XLNET

    else:
        if model in ['albert', 'bert']:
            selected_class = BINARY_BERT
        if model in ['xlnet']:
            selected_class = BINARY_XLNET

    if model in ['albert', 'bert']:
        if model == 'bert':
            from .._transformer._bert import _extract_attention_weights_import
        if model == 'albert':
            from .._transformer._albert import _extract_attention_weights_import

        tokenizer, cls, sep = sentencepiece_tokenizer_bert(
            path[model][size]['tokenizer'], path[model][size]['vocab']
        )

        return selected_class(
            X = g.get_tensor_by_name('import/Placeholder:0'),
            segment_ids = None,
            input_masks = None,
            logits = g.get_tensor_by_name('import/logits:0'),
            logits_seq = g.get_tensor_by_name('import/logits_seq:0'),
            sess = generate_session(graph = g),
            tokenizer = tokenizer,
            label = label,
            cls = cls,
            sep = sep,
            attns = _extract_attention_weights_import(bert_num_layers[size], g),
            class_name = class_name,
        )
    if model in ['xlnet']:
        from .._transformer._xlnet import _extract_attention_weights_import

        tokenizer = sentencepiece_tokenizer_xlnet(
            path[model][size]['tokenizer']
        )

        return selected_class(
            X = g.get_tensor_by_name('import/Placeholder:0'),
            segment_ids = g.get_tensor_by_name('import/Placeholder_1:0'),
            input_masks = g.get_tensor_by_name('import/Placeholder_2:0'),
            logits = g.get_tensor_by_name('import/logits:0'),
            logits_seq = g.get_tensor_by_name('import/logits_seq:0'),
            sess = generate_session(graph = g),
            tokenizer = tokenizer,
            label = label,
            attns = _extract_attention_weights_import(g),
            class_name = class_name,
        )



from sklearn.metrics.pairwise import cosine_similarity
from sklearn.cluster import KMeans
from sklearn.manifold import MDS
from sklearn.feature_extraction.text import TfidfVectorizer, CountVectorizer
from .texts.vectorizer import SkipGramVectorizer
from .stem import sastrawi
from .texts._text_functions import (
    simple_textcleaning,
    split_into_sentences,
    STOPWORDS,
)

import numpy as np
import re
import random

_accepted_pos = [
    'ADJ',
    'ADP',
    'ADV',
    'AUX',
    'CCONJ',
    'DET',
    'NOUN',
    'NUM',
    'PART',
    'PRON',
    'PROPN',
    'SCONJ',
    'SYM',
    'VERB',
    'X',
]
_accepted_entities = [
    'OTHER',
    'law',
    'location',
    'organization',
    'person',
    'quantity',
    'time',
    'event',
    'X',
]


def cluster_words(list_words):
    """
    cluster similar words based on structure, eg, ['mahathir mohamad', 'mahathir'] = ['mahathir mohamad']

    Parameters
    ----------
    list_words : list of str

    Returns
    -------
    string: list of clustered words
    """
    if not isinstance(list_words, list):
        raise ValueError('list_words must be a list')

    if not len(list_words):
        return []

    if not isinstance(list_words[0], str):
        raise ValueError('list_words must be a list of strings')

    dict_words = {}
    for word in list_words:
        found = False
        for key in dict_words.keys():
            if word in key or any(
                [word in inside for inside in dict_words[key]]
            ):
                dict_words[key].append(word)
                found = True
            if key in word:
                dict_words[key].append(word)
        if not found:
            dict_words[word] = [word]
    results = []
    for key, words in dict_words.items():
        results.append(max(list(set([key] + words)), key = len))
    return list(set(results))


def cluster_pos(result):
    """
    cluster similar POS.

    Parameters
    ----------
    result: list

    Returns
    -------
    result: list
    """
    if not isinstance(result, list):
        raise ValueError('result must be a list')
    if not isinstance(result[0], tuple):
        raise ValueError('result must be a list of tuple')
    if not all([i[1] in _accepted_pos for i in result]):
        raise ValueError(
            'elements of result must be a subset or equal of supported POS, please run malaya.describe_pos() to get supported POS'
        )

    output = {
        'ADJ': [],
        'ADP': [],
        'ADV': [],
        'ADX': [],
        'AUX': [],
        'CCONJ': [],
        'DET': [],
        'NOUN': [],
        'NUM': [],
        'PART': [],
        'PRON': [],
        'PROPN': [],
        'SCONJ': [],
        'SYM': [],
        'VERB': [],
        'X': [],
    }
    last_label, words = None, []
    for word, label in result:
        if last_label != label and last_label:
            joined = ' '.join(words)
            if joined not in output[last_label]:
                output[last_label].append(joined)
            words = []
            last_label = label
            words.append(word)

        else:
            if not last_label:
                last_label = label
            words.append(word)
    output[last_label].append(' '.join(words))
    return output


def cluster_tagging(result):
    """
    cluster any tagging results, as long the data passed `[(string, label), (string, label)]`.

    Parameters
    ----------
    result: list

    Returns
    -------
    result: list
    """
    if not isinstance(result, list):
        raise ValueError('result must be a list')
    if not isinstance(result[0], tuple):
        raise ValueError('result must be a list of tuple')

    _, labels = list(zip(*result))

    output = {l: [] for l in labels}
    last_label, words = None, []
    for word, label in result:
        if last_label != label and last_label:
            joined = ' '.join(words)
            if joined not in output[last_label]:
                output[last_label].append(joined)
            words = []
            last_label = label
            words.append(word)

        else:
            if not last_label:
                last_label = label
            words.append(word)
    output[last_label].append(' '.join(words))
    return output


def cluster_entities(result):
    """
    cluster similar Entities.

    Parameters
    ----------
    result: list

    Returns
    -------
    result: list
    """
    if not isinstance(result, list):
        raise ValueError('result must be a list')
    if not isinstance(result[0], tuple):
        raise ValueError('result must be a list of tuple')
    if not all([i[1] in _accepted_entities for i in result]):
        raise ValueError(
            'elements of result must be a subset or equal of supported Entities, please run malaya.describe_entities() to get supported POS'
        )

    output = {
        'OTHER': [],
        'law': [],
        'location': [],
        'organization': [],
        'person': [],
        'quantity': [],
        'time': [],
        'event': [],
        'X': [],
    }
    last_label, words = None, []
    for word, label in result:
        if last_label != label and last_label:
            joined = ' '.join(words)
            if joined not in output[last_label]:
                output[last_label].append(joined)
            words = []
            last_label = label
            words.append(word)

        else:
            if not last_label:
                last_label = label
            words.append(word)
    output[last_label].append(' '.join(words))
    return output


def cluster_scatter(
    corpus,
    vectorizer,
    num_clusters = 5,
    titles = None,
    colors = None,
    stemming = True,
    stop_words = None,
    cleaning = simple_textcleaning,
    clustering = KMeans,
    decomposition = MDS,
    ngram = (1, 3),
    figsize = (17, 9),
    batch_size = 20,
):
    """
    plot scatter plot on similar text clusters.

    Parameters
    ----------

    corpus: list
    vectorizer: class
    num_clusters: int, (default=5)
        size of unsupervised clusters.
    titles: list
        list of titles, length must same with corpus.
    colors: list
        list of colors, length must same with num_clusters.
    stemming: bool, (default=True)
        If True, sastrawi_stemmer will apply.
    stop_words: list, (default=None)
        list of stop words to remove. If None, default is malaya.texts._text_functions.STOPWORDS
    ngram: tuple, (default=(1,3))
        n-grams size to train a corpus.
    cleaning: function, (default=simple_textcleaning)
        function to clean the corpus.
    batch_size: int, (default=10)
        size of strings for each vectorization and attention. Only useful if use transformer vectorizer.

    Returns
    -------
    dictionary: {'X': X, 'Y': Y, 'labels': clusters, 'vector': transformed_text_clean, 'titles': titles}
    """
    if not isinstance(corpus, list):
        raise ValueError('corpus must be a list')
    if not isinstance(corpus[0], str):
        raise ValueError('corpus must be list of strings')
    if not isinstance(titles, list) and titles is not None:
        raise ValueError('titles must be a list or None')
    if not isinstance(colors, list) and colors is not None:
        raise ValueError('colors must be a list or None')
    if not isinstance(batch_size, int):
        raise ValueError('batch_size must be an integer')
    if titles:
        if len(titles) != len(corpus):
            raise ValueError('length of titles must be same with corpus')
    if colors:
        if len(colors) != num_clusters:
            raise ValueError(
                'size of colors must be same with number of clusters'
            )
    if not hasattr(vectorizer, 'vectorize') and not hasattr(vectorizer, 'fit'):
        raise ValueError('vectorizer must has `fit` and `vectorize` methods')
    if not isinstance(stemming, bool):
        raise ValueError('bool must be a boolean')

    try:
        import matplotlib.pyplot as plt
        import seaborn as sns

        sns.set()
    except:
        raise Exception(
            'matplotlib and seaborn not installed. Please install it and try again.'
        )

    if stop_words is None:
        stop_words = STOPWORDS

    if cleaning is not None:
        for i in range(len(corpus)):
            corpus[i] = cleaning(corpus[i])
    if stemming:
        for i in range(len(corpus)):
            corpus[i] = sastrawi(corpus[i])
    text_clean = []
    for text in corpus:
        text_clean.append(
            ' '.join([word for word in text.split() if word not in stop_words])
        )

    if hasattr(vectorizer, 'fit'):
        vectorizer.fit(text_clean)
        transformed_text_clean = vectorizer.transform(text_clean)
        features = vectorizer.get_feature_names()
    else:
        transformed_text_clean, attentions = [], []
        for i in range(0, len(text_clean), batch_size):
            index = min(i + batch_size, len(text_clean))
            transformed_text_clean.append(
                vectorizer.vectorize(text_clean[i:index])
            )
            attentions.extend(vectorizer.attention(text_clean[i:index]))
        transformed_text_clean = np.concatenate(
            transformed_text_clean, axis = 0
        )
    km = clustering(n_clusters = num_clusters)
    dist = 1 - cosine_similarity(transformed_text_clean)
    km.fit(transformed_text_clean)
    clusters = km.labels_.tolist()
    if isinstance(decomposition, MDS):
        decomposed = decomposition(
            n_components = 2, dissimilarity = 'precomputed'
        )
    else:
        decomposed = decomposition(n_components = 2)
    pos = decomposed.fit_transform(dist)
    if not titles:
        titles = []
        for i in range(transformed_text_clean.shape[0]):

            if hasattr(vectorizer, 'fit'):
                indices = np.argsort(
                    np.array(transformed_text_clean[i].todense())[0]
                )[::-1]
                titles.append(
                    ' '.join([features[i] for i in indices[: ngram[1]]])
                )
            else:
                attentions[i].sort(key = lambda x: x[1])
                titles.append(
                    ' '.join([i[0] for i in attentions[i][-ngram[1] :]])
                )

    if not colors:
        colors = sns.color_palette(n_colors = num_clusters)
    X, Y = pos[:, 0], pos[:, 1]
    plt.figure(figsize = figsize)
    for i in np.unique(clusters):
        plt.scatter(
            X[clusters == i],
            Y[clusters == i],
            color = colors[i],
            label = 'cluster %d' % (i),
        )
    for i in range(len(X)):
        plt.text(X[i], Y[i], titles[i], size = 8)
    plt.legend()
    plt.show()
    return {
        'X': X,
        'Y': Y,
        'labels': clusters,
        'vector': transformed_text_clean,
        'titles': titles,
    }


def cluster_dendogram(
    corpus,
    vectorizer,
    titles = None,
    stemming = True,
    stop_words = None,
    cleaning = simple_textcleaning,
    random_samples = 0.3,
    ngram = (1, 3),
    figsize = (17, 9),
    batch_size = 20,
    **kwargs
):
    """
    plot hierarchical dendogram with similar texts.

    Parameters
    ----------

    corpus: list
    vectorizer: class
    num_clusters: int, (default=5)
        size of unsupervised clusters.
    titles: list
        list of titles, length must same with corpus.
    stemming: bool, (default=True)
        If True, sastrawi_stemmer will apply.
    stop_words: list, (default=None)
        list of stop words to remove. If None, default is malaya.texts._text_functions.STOPWORDS
    cleaning: function, (default=simple_textcleaning)
        function to clean the corpus.
    random_samples: float, (default=0.3)
        random samples from the corpus, 0.3 means 30%.
    ngram: tuple, (default=(1,3))
        n-grams size to train a corpus.
    batch_size: int, (default=20)
        size of strings for each vectorization and attention. Only useful if use transformer vectorizer.

    Returns
    -------
    dictionary: {'linkage_matrix': linkage_matrix, 'titles': titles}
    """
    if not isinstance(corpus, list):
        raise ValueError('corpus must be a list')
    if not isinstance(corpus[0], str):
        raise ValueError('corpus must be list of strings')
    if not isinstance(titles, list) and titles is not None:
        raise ValueError('titles must be a list or None')
    if titles:
        if len(titles) != len(corpus):
            raise ValueError('length of titles must be same with corpus')
    if not isinstance(stemming, bool):
        raise ValueError('bool must be a boolean')

    if not hasattr(vectorizer, 'vectorize') and not hasattr(vectorizer, 'fit'):
        raise ValueError('vectorizer must has `fit` and `vectorize` methods')
    if not isinstance(random_samples, float):
        raise ValueError('random_samples must be a float')
    if not (random_samples < 1 and random_samples > 0):
        raise ValueError('random_samples must be between 0 and 1')
    if not isinstance(ngram, tuple):
        raise ValueError('ngram must be a tuple')
    if not len(ngram) == 2:
        raise ValueError('ngram size must equal to 2')

    try:
        import matplotlib.pyplot as plt
        import seaborn as sns
        from scipy.cluster.hierarchy import ward, dendrogram

        sns.set()
    except:
        raise Exception(
            'matplotlib and seaborn not installed. Please install it and try again.'
        )
    if stop_words is None:
        stop_words = STOPWORDS

    corpus = random.sample(corpus, k = int(random_samples * len(corpus)))

    if cleaning is not None:
        for i in range(len(corpus)):
            corpus[i] = cleaning(corpus[i])
    if stemming:
        for i in range(len(corpus)):
            corpus[i] = sastrawi(corpus[i])
    text_clean = []
    for text in corpus:
        text_clean.append(
            ' '.join([word for word in text.split() if word not in stop_words])
        )

    if hasattr(vectorizer, 'fit'):
        vectorizer.fit(text_clean)
        transformed_text_clean = vectorizer.transform(text_clean)
        features = vectorizer.get_feature_names()
    else:
        transformed_text_clean, attentions = [], []
        for i in range(0, len(text_clean), batch_size):
            index = min(i + batch_size, len(text_clean))
            transformed_text_clean.append(
                vectorizer.vectorize(text_clean[i:index])
            )
            attentions.extend(vectorizer.attention(text_clean[i:index]))
        transformed_text_clean = np.concatenate(
            transformed_text_clean, axis = 0
        )

    dist = 1 - cosine_similarity(transformed_text_clean)
    linkage_matrix = ward(dist)
    if not titles:
        titles = []
        for i in range(transformed_text_clean.shape[0]):

            if hasattr(vectorizer, 'fit'):
                indices = np.argsort(
                    np.array(transformed_text_clean[i].todense())[0]
                )[::-1]
                titles.append(
                    ' '.join([features[i] for i in indices[: ngram[1]]])
                )
            else:
                attentions[i].sort(key = lambda x: x[1])
                titles.append(
                    ' '.join([i[0] for i in attentions[i][-ngram[1] :]])
                )
    plt.figure(figsize = figsize)
    ax = dendrogram(linkage_matrix, orientation = 'right', labels = titles)
    plt.tick_params(
        axis = 'x',
        which = 'both',
        bottom = 'off',
        top = 'off',
        labelbottom = 'off',
    )
    plt.tight_layout()
    plt.show()
    return {'linkage_matrix': linkage_matrix, 'titles': titles}


def cluster_graph(
    corpus,
    vectorizer,
    threshold = 0.9,
    num_clusters = 5,
    titles = None,
    colors = None,
    stop_words = None,
    stemming = True,
    ngram = (1, 3),
    cleaning = simple_textcleaning,
    clustering = KMeans,
    figsize = (17, 9),
    with_labels = True,
    batch_size = 20,
    **kwargs
):
    """
    plot undirected graph with similar texts.

    Parameters
    ----------

    corpus: list
    vectorizer: class
    threshold: float, (default=0.9)
        0.9 means, 90% above absolute pearson correlation.
    num_clusters: int, (default=5)
        size of unsupervised clusters.
    titles: list
        list of titles, length must same with corpus.
    stemming: bool, (default=True)
        If True, sastrawi_stemmer will apply.
    stop_words: list, (default=None)
        list of stop words to remove. If None, default is malaya.texts._text_functions.STOPWORDS
    cleaning: function, (default=simple_textcleaning)
        function to clean the corpus.
    ngram: tuple, (default=(1,3))
        n-grams size to train a corpus.
    batch_size: int, (default=20)
        size of strings for each vectorization and attention. Only useful if use transformer vectorizer.

    Returns
    -------
    dictionary: {'G': G, 'pos': pos, 'node_colors': node_colors, 'node_labels': node_labels}
    """

    if not isinstance(corpus, list):
        raise ValueError('corpus must be a list')
    if not isinstance(corpus[0], str):
        raise ValueError('corpus must be list of strings')
    if not isinstance(titles, list) and titles is not None:
        raise ValueError('titles must be a list or None')
    if not isinstance(colors, list) and colors is not None:
        raise ValueError('colors must be a list or None')
    if titles:
        if len(titles) != len(corpus):
            raise ValueError('length of titles must be same with corpus')
    if colors:
        if len(colors) != num_clusters:
            raise ValueError(
                'size of colors must be same with number of clusters'
            )
    if not hasattr(vectorizer, 'vectorize') and not hasattr(vectorizer, 'fit'):
        raise ValueError('vectorizer must has `fit` and `vectorize` methods')
    if not isinstance(stemming, bool):
        raise ValueError('bool must be a boolean')
    if not isinstance(ngram, tuple):
        raise ValueError('ngram must be a tuple')
    if not len(ngram) == 2:
        raise ValueError('ngram size must equal to 2')
    if not isinstance(threshold, float):
        raise ValueError('threshold must be a float')
    if not (threshold <= 1 and threshold > 0):
        raise ValueError(
            'threshold must be bigger than 0, less than or equal to 1'
        )

    try:
        import matplotlib.pyplot as plt
        import seaborn as sns
        import networkx as nx
        import networkx.drawing.layout as nxlayout
        import pandas as pd

        sns.set()
    except:
        raise Exception(
            'matplotlib, seaborn, networkx not installed. Please install it and try again.'
        )
    if stop_words is None:
        stop_words = STOPWORDS

    if cleaning is not None:
        for i in range(len(corpus)):
            corpus[i] = cleaning(corpus[i])
    if stemming:
        for i in range(len(corpus)):
            corpus[i] = sastrawi(corpus[i])
    text_clean = []
    for text in corpus:
        text_clean.append(
            ' '.join([word for word in text.split() if word not in stop_words])
        )

    if hasattr(vectorizer, 'fit'):
        vectorizer.fit(text_clean)
        transformed_text_clean = vectorizer.transform(text_clean).todense()
        features = vectorizer.get_feature_names()
    else:
        transformed_text_clean, attentions = [], []
        for i in range(0, len(text_clean), batch_size):
            index = min(i + batch_size, len(text_clean))
            transformed_text_clean.append(
                vectorizer.vectorize(text_clean[i:index])
            )
            attentions.extend(vectorizer.attention(text_clean[i:index]))
        transformed_text_clean = np.concatenate(
            transformed_text_clean, axis = 0
        )

    DxT = transformed_text_clean
    DxD = np.abs(pd.DataFrame(DxT.T).corr()).values
    km = clustering(n_clusters = num_clusters)
    km.fit(DxT)
    clusters = km.labels_.tolist()

    if not titles:
        titles = []
        for i in range(transformed_text_clean.shape[0]):

            if hasattr(vectorizer, 'fit'):
                indices = np.argsort(np.array(transformed_text_clean[i])[0])[
                    ::-1
                ]
                titles.append(
                    ' '.join([features[i] for i in indices[: ngram[1]]])
                )
            else:
                attentions[i].sort(key = lambda x: x[1])
                titles.append(
                    ' '.join([i[0] for i in attentions[i][-ngram[1] :]])
                )

    if not colors:
        colors = sns.color_palette(n_colors = num_clusters)
    G = nx.Graph()
    for i in range(DxT.shape[0]):
        G.add_node(i, text = titles[i], label = clusters[i])

    len_dense = len(DxD)
    for i in range(len_dense):
        for j in range(len_dense):
            if j == i:
                continue
            if DxD[i, j] >= threshold:
                weight = DxD[i, j]
                G.add_edge(i, j, weight = weight)
    node_colors, node_labels = [], {}
    for node in G:
        node_colors.append(colors[G.node[node]['label']])
        node_labels[node] = G.node[node]['text']
    pos = nxlayout.fruchterman_reingold_layout(
        G, k = 1.5 / np.sqrt(len(G.nodes()))
    )
    plt.figure(figsize = figsize)
    if with_labels:
        nx.draw(G, node_color = node_colors, pos = pos, labels = node_labels)
    else:
        nx.draw(G, node_color = node_colors, pos = pos)

    return {
        'G': G,
        'pos': pos,
        'node_colors': node_colors,
        'node_labels': node_labels,
    }


def cluster_entity_linking(
    corpus,
    vectorizer,
    entity_model,
    topic_modeling_model,
    threshold = 0.3,
    topic_decomposition = 2,
    topic_length = 10,
    fuzzy_ratio = 70,
    accepted_entities = ['law', 'location', 'organization', 'person', 'event'],
    cleaning = simple_textcleaning,
    stemming = True,
    colors = None,
    stop_words = None,
    max_df = 1.0,
    min_df = 1,
    ngram = (2, 3),
    figsize = (17, 9),
    batch_size = 20,
    **kwargs
):
    """
    plot undirected graph for Entities and topics relationship.

    Parameters
    ----------
    corpus: list or str
    vectorizer: class
    titles: list
        list of titles, length must same with corpus.
    colors: list
        list of colors, length must same with num_clusters.
    threshold: float, (default=0.3)
        0.3 means, 30% above absolute pearson correlation.
    topic_decomposition: int, (default=2)
        size of decomposition.
    topic_length: int, (default=10)
        size of topic models.
    fuzzy_ratio: int, (default=70)
        size of ratio for fuzzywuzzy.
    stemming: bool, (default=True)
        If True, sastrawi_stemmer will apply.
    max_df: float, (default=0.95)
        maximum of a word selected based on document frequency.
    min_df: int, (default=2)
        minimum of a word selected on based on document frequency.
    ngram: tuple, (default=(1,3))
        n-grams size to train a corpus.
    cleaning: function, (default=simple_textcleaning)
        function to clean the corpus.
    stop_words: list, (default=None)
        list of stop words to remove. If None, default is malaya.texts._text_functions.STOPWORDS

    Returns
    -------
    dictionary: {'G': G, 'pos': pos, 'node_colors': node_colors, 'node_labels': node_labels}
    """

    import inspect

    if not isinstance(corpus, list) and not isinstance(corpus, str):
        raise ValueError('corpus must be a list')
    if isinstance(corpus, list):
        if not isinstance(corpus[0], str):
            raise ValueError('corpus must be list of strings')
    if not hasattr(vectorizer, 'vectorize') and not hasattr(vectorizer, 'fit'):
        raise ValueError('vectorizer must has `fit` and `vectorize` methods')
    if 'max_df' not in inspect.getargspec(topic_modeling_model)[0]:
        raise ValueError('topic_modeling_model must has `max_df` parameter')
    if not isinstance(colors, list) and colors is not None:
        raise ValueError('colors must be a list or None')
    if not isinstance(stemming, bool):
        raise ValueError('bool must be a boolean')
    if not isinstance(ngram, tuple):
        raise ValueError('ngram must be a tuple')
    if not len(ngram) == 2:
        raise ValueError('ngram size must equal to 2')
    if not isinstance(min_df, int):
        raise ValueError('min_df must be an integer')
    if not isinstance(topic_decomposition, int):
        raise ValueError('topic_decomposition must be an integer')
    if not isinstance(topic_length, int):
        raise ValueError('topic_length must be an integer')
    if not isinstance(fuzzy_ratio, int):
        raise ValueError('fuzzy_ratio must be an integer')
    if not isinstance(max_df, float):
        raise ValueError('max_df must be a float')

    if min_df < 1:
        raise ValueError('min_df must be bigger than 0')
    if not (max_df <= 1 and max_df > 0):
        raise ValueError(
            'max_df must be bigger than 0, less than or equal to 1'
        )
    if not (fuzzy_ratio > 0 and fuzzy_ratio <= 100):
        raise ValueError(
            'fuzzy_ratio must be bigger than 0, less than or equal to 100'
        )
    if not isinstance(threshold, float):
        raise ValueError('threshold must be a float')
    if not (threshold <= 1 and threshold > 0):
        raise ValueError(
            'threshold must be bigger than 0, less than or equal to 1'
        )
    if stop_words is None:
        stop_words = STOPWORDS

    try:
        import matplotlib.pyplot as plt
        import seaborn as sns
        import networkx as nx
        import networkx.drawing.layout as nxlayout
        import pandas as pd
        from fuzzywuzzy import fuzz

        sns.set()
    except:
        raise Exception(
            'matplotlib, seaborn, networkx, fuzzywuzzy not installed. Please install it and try again.'
        )

    if isinstance(corpus, str):
        corpus = split_into_sentences(corpus)
    else:
        corpus = '. '.join(corpus)
        corpus = split_into_sentences(corpus)

    corpus = [string for string in corpus if len(string) > 5]

    if not colors:
        colors = sns.color_palette(n_colors = len(accepted_entities) + 1)
    else:
        if len(colors) != (len(accepted_entities) + 1):
            raise ValueError(
                'len of colors must same as %d' % (len(accepted_entities) + 1)
            )

    topic_model = topic_modeling_model(
        corpus,
        topic_decomposition,
        ngram = ngram,
        max_df = max_df,
        min_df = min_df,
    )
    topics = []
    for no, topic in enumerate(topic_model.comp.components_):
        for i in topic.argsort()[: -topic_length - 1 : -1]:
            topics.append(topic_model.features[i])

    entities_cluster = {entity: [] for entity in accepted_entities}
    for string in corpus:
        entities_clustered = cluster_entities(entity_model.predict(string))
        for entity in accepted_entities:
            entities_cluster[entity].extend(entities_clustered[entity])
    for entity in accepted_entities:
        entities_cluster[entity] = cluster_words(
            list(set(entities_cluster[entity]))
        )

    topics = cluster_words(list(set(topics)))
    color_dict = {topic: colors[-1] for topic in topics}
    for no, entity in enumerate(accepted_entities):
        for e in entities_cluster[entity]:
            topics.append(e)
            color_dict[e] = colors[no]

    topics_corpus = []
    for topic in topics:
        nested_corpus = []
        for string in corpus:
            if (
                topic in string
                or fuzz.token_set_ratio(topic, string) >= fuzzy_ratio
            ):
                nested_corpus.append(string)
        topics_corpus.append(' '.join(nested_corpus))

    corpus = topics_corpus

    if cleaning is not None:
        for i in range(len(corpus)):
            corpus[i] = cleaning(corpus[i])
    if stemming:
        for i in range(len(corpus)):
            corpus[i] = sastrawi(corpus[i])
    text_clean = []
    for text in corpus:
        text_clean.append(
            ' '.join([word for word in text.split() if word not in stop_words])
        )

    if hasattr(vectorizer, 'fit'):
        vectorizer.fit(text_clean)
        transformed_text_clean = vectorizer.transform(text_clean).todense()
        features = vectorizer.get_feature_names()
    else:
        transformed_text_clean, attentions = [], []
        for i in range(0, len(text_clean), batch_size):
            index = min(i + batch_size, len(text_clean))
            transformed_text_clean.append(
                vectorizer.vectorize(text_clean[i:index])
            )
            attentions.extend(vectorizer.attention(text_clean[i:index]))
        transformed_text_clean = np.concatenate(
            transformed_text_clean, axis = 0
        )

    DxT = transformed_text_clean
    DxD = np.abs(pd.DataFrame(DxT.T).corr()).values

    G = nx.Graph()
    for i in range(DxT.shape[0]):
        G.add_node(i, text = topics[i], label = topics[i])

    len_dense = len(DxD)
    for i in range(len_dense):
        for j in range(len_dense):
            if j == i:
                continue
            if DxD[i, j] >= threshold:
                weight = DxD[i, j]
                G.add_edge(i, j, weight = weight)

    node_colors, node_labels = [], {}
    for node in G:
        node_colors.append(color_dict[G.node[node]['label']])
        node_labels[node] = G.node[node]['text']
    pos = nxlayout.fruchterman_reingold_layout(
        G, k = 1.5 / np.sqrt(len(G.nodes()))
    )
    f = plt.figure(figsize = figsize)
    ax = f.add_subplot(1, 1, 1)
    for no, entity in enumerate(accepted_entities):
        ax.plot([0], [0], color = colors[no], label = entity)
    ax.plot([0], [0], color = colors[-1], label = 'topics')
    nx.draw(
        G, node_color = node_colors, pos = pos, labels = node_labels, ax = ax
    )
    plt.legend()
    plt.tight_layout()
    plt.show()
    return {
        'G': G,
        'pos': pos,
        'node_colors': node_colors,
        'node_labels': node_labels,
    }



_availability = {
    'bert': ['base', 'small'],
    'xlnet': ['base'],
    'alxlnet': ['base'],
    'albert': ['base', 'large'],
}


def available_model():
    """
    List available transformer models.
    """
    return _availability


def load(size = 'base', model = 'xlnet', pool_mode = 'last', validate = True):

    """
    Load transformer model.

    Parameters
    ----------
    model : str, optional (default='bert')
        Model architecture supported. Allowed values:

        * ``'bert'`` - BERT architecture from google.
        * ``'xlnet'`` - XLNET architecture from google.
        * ``'alxlnet'`` - XLNET architecture from google + Malaya.
        * ``'albert'`` - ALBERT architecture from google.
    size : str, optional (default='base')
        Model size supported. Allowed values:

        * ``'base'`` - BASE size.
        * ``'small'`` - SMALL size.
        * ``'large'`` - LARGE size.
    pool_mode : str, optional (default='last')
        Model logits architecture supported. Only usable if model = 'xlnet'. Allowed values:

        * ``'last'`` - last of the sequence.
        * ``'first'`` - first of the sequence.
        * ``'mean'`` - mean of the sequence.
        * ``'attn'`` - attention of the sequence.
    validate: bool, optional (default=True)
        if True, malaya will check model availability and download if not available.

    Returns
    -------
    TRANSFORMER: malaya.transformer class
    """

    if not isinstance(model, str):
        raise ValueError('model must be a string')
    if not isinstance(size, str):
        raise ValueError('size must be a string')
    if not isinstance(pool_mode, str):
        raise ValueError('pool_mode must be a string')
    if not isinstance(validate, bool):
        raise ValueError('validate must be a boolean')

    model = model.lower()
    size = size.lower()
    if model not in _availability:
        raise Exception(
            'model not supported, please check supported models from malaya.transformer.available_model()'
        )
    if size not in _availability[model]:
        raise Exception(
            'size not supported, please check supported models from malaya.transformer.available_model()'
        )
    if model == 'bert':
        from ._transformer import _bert

        return _bert.bert(model = size, validate = validate)
    if model == 'albert':
        from ._transformer import _albert

        return _albert.albert(model = size, validate = validate)
    if model == 'xlnet':
        from ._transformer import _xlnet

        return _xlnet.xlnet(
            model = size, pool_mode = pool_mode, validate = validate
        )
    if model == 'alxlnet':
        from ._transformer import _alxlnet

        return _alxlnet.alxlnet(
            model = size, pool_mode = pool_mode, validate = validate
        )



# Malaya Natural Language Toolkit
#
# Copyright (C) 2019 Malaya Project
# Licensed under the MIT License
# Author: huseinzol05 <husein.zol05@gmail.com>
# URL: <https://malaya.readthedocs.io/>
# For license information, see https://github.com/huseinzol05/Malaya/blob/master/LICENSE

from ._tatabahasa import (
    tanya_list,
    perintah_list,
    pangkal_list,
    bantu_list,
    penguat_list,
    penegas_list,
    nafi_list,
    pemeri_list,
    sendi_list,
    pembenar_list,
    nombor_list,
    suku_bilangan_list,
    pisahan_list,
    keterangan_list,
    arah_list,
    hubung_list,
    gantinama_list,
    permulaan,
    hujung,
    hujung_malaysian,
    calon_dictionary,
    stopwords_calon,
)
from ._text_functions import STOPWORDS as stopwords
from ._text_functions import split_into_sentences as sentence_splitter
